# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import tobiko
from tobiko import config
from tobiko.openstack import glance
from tobiko.openstack import topology
from tobiko.openstack.stacks import _nova
from tobiko.openstack.stacks import _vlan
from tobiko.shell import sh


CONF = config.CONF


class UbuntuMinimalImageFixture(glance.FileGlanceImageFixture):
    image_name = CONF.tobiko.ubuntu.image_name
    image_file = CONF.tobiko.ubuntu.image_file
    disk_format = CONF.tobiko.ubuntu.disk_format or "qcow2"
    container_format = CONF.tobiko.ubuntu.container_format or "bare"
    username = CONF.tobiko.ubuntu.username or 'ubuntu'
    password = CONF.tobiko.ubuntu.password or 'ubuntu'
    connection_timeout = CONF.tobiko.nova.ubuntu_connection_timeout
    disabled_algorithms = CONF.tobiko.ubuntu.disabled_algorithms
    is_reachable_timeout = CONF.tobiko.nova.ubuntu_is_reachable_timeout


IPERF3_SERVICE_FILE = """
[Unit]
Description=iperf3 server on port %i
After=syslog.target network.target

[Service]
ExecStart=/usr/bin/iperf3 -s -p %i
Restart=always
User=root

[Install]
WantedBy=multi-user.target
DefaultInstance=5201
"""


class UbuntuImageFixture(UbuntuMinimalImageFixture,
                         glance.CustomizedGlanceImageFixture):
    """Ubuntu server image running an HTTP server

    The server has additional installed packages compared to
    the minimal one:

    - iperf3
    - ping
    - ncat
    - nginx
    - vlan

    The image will also have below running services:

    - nginx HTTP server listening on TCP port 80
    - iperf3 server listening on TCP port 5201
    """

    def __init__(self,
                 ethernet_devide: str = None,
                 **kwargs):
        super().__init__(**kwargs)
        self._ethernet_device = ethernet_devide

    @property
    def firstboot_commands(self) -> typing.List[str]:
        return super().firstboot_commands + [
            'sh -c "hostname > /var/www/html/id"']

    @property
    def install_packages(self) -> typing.List[str]:
        return super().install_packages + ['iperf3',
                                           'iputils-ping',
                                           'nano',
                                           'ncat',
                                           'nginx',
                                           'vlan']

    # port of running HTTP server
    http_port = 80

    # port of running Iperf3 server
    iperf3_port = 5201

    @property
    def iperf3_service_name(self) -> str:
        return f"iperf3-server@{self.iperf3_port}.service"

    @property
    def run_commands(self) -> typing.List[str]:
        run_commands = super().run_commands
        run_commands.append(
            f'echo "{IPERF3_SERVICE_FILE}" '
            '> /etc/systemd/system/iperf3-server@.service')
        run_commands.append(
            f"systemctl enable iperf3-server@{self.iperf3_port}")
        run_commands.append('echo "8021q" >> /etc/modules')
        return run_commands

    @property
    def ethernet_device(self) -> str:
        if self._ethernet_device is None:
            self._ethernet_device = self._get_ethernet_device()
        return self._ethernet_device

    @staticmethod
    def _get_ethernet_device() -> str:
        """From OSP17 and above, Ubuntu stack should use a different interface name.
        This method returns the interface name, depending on the OSP version.
        """
        if_name = CONF.tobiko.ubuntu.interface_name
        if if_name is not None:
            return if_name
        from tobiko import tripleo
        if tripleo.has_overcloud(min_version='17.0'):
            return 'enp3s0'
        else:
            return 'ens3'

    def _get_customized_suffix(self) -> str:
        return f'{super()._get_customized_suffix()}-{self.ethernet_device}'

    @property
    def vlan_id(self) -> int:
        return tobiko.tobiko_config().neutron.vlan_id

    @property
    def vlan_device(self) -> str:
        return f'vlan{self.vlan_id}'

    @property
    def vlan_config(self) -> typing.Dict[str, typing.Any]:
        return {
            'network': {
                'version': 2,
                'vlans': {
                    self.vlan_device: {
                        'link': self.ethernet_device,
                        'dhcp4': True,
                        'dhcp4-overrides': {
                            'use-routes': False
                        },
                        'dhcp6': True,
                        'dhcp6-overrides': {
                            'use-routes': False
                        },
                        'id': self.vlan_id,
                    }
                }
            }
        }

    @property
    def write_files(self) -> typing.Dict[str, str]:
        write_files = super().write_files
        write_files['/etc/netplan/75-tobiko-vlan.yaml'] = tobiko.dump_yaml(
            self.vlan_config)
        return write_files


class UbuntuFlavorStackFixture(_nova.FlavorStackFixture):
    ram = 256
    swap = 512


class UbuntuMinimalServerStackFixture(_nova.CloudInitServerStackFixture):

    #: Glance image used to create a Nova server instance
    image_fixture = tobiko.required_fixture(UbuntuMinimalImageFixture)

    #: Flavor used to create a Nova server instance
    flavor_stack = tobiko.required_fixture(UbuntuFlavorStackFixture)


class UbuntuServerStackFixture(UbuntuMinimalServerStackFixture,
                               _vlan.VlanServerStackFixture):
    """Ubuntu server running an HTTP server

    The server has additional commands compared to the minimal one:

    - iperf3
    - ping
    """

    #: Glance image used to create a Nova server instance
    image_fixture = tobiko.required_fixture(UbuntuImageFixture)

    @property
    def is_reachable_timeout(self) -> tobiko.Seconds:
        return self.image_fixture.is_reachable_timeout

    # port of running HTTP server
    @property
    def http_port(self) -> int:
        return self.image_fixture.http_port

    @property
    def iperf3_port(self) -> int:
        return self.image_fixture.iperf3_port

    @property
    def iperf3_service_name(self) -> str:
        return self.image_fixture.iperf3_service_name

    def wait_for_iperf3_server(self,
                               timeout: tobiko.Seconds = None,
                               interval: tobiko.Seconds = None):
        return sh.wait_for_active_systemd_units(self.iperf3_service_name,
                                                timeout=timeout,
                                                interval=interval,
                                                ssh_client=self.ssh_client)

    @property
    def vlan_id(self) -> int:
        return self.image_fixture.vlan_id

    @property
    def vlan_device(self) -> str:
        return self.image_fixture.vlan_device


@topology.skip_unless_osp_version('17.0', lower=True)
class UbuntuExternalServerStackFixture(UbuntuServerStackFixture,
                                       _nova.ExternalServerStackFixture):
    """Ubuntu server with port on special external network
    """



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack.openstackclient import _client


def network_loggable_resources_list(*args, **kwargs):
    cmd = 'openstack network loggable resources list {params}'
    kwargs['format'] = 'json'
    return _client.execute(cmd, *args, **kwargs)


def network_log_create(log_name, *args, **kwargs):
    cmd = f'openstack network log create {{params}} {log_name}'
    kwargs['format'] = 'json'
    return _client.execute(cmd, *args, **kwargs)


def network_log_show(log_name, *args, **kwargs):
    cmd = f'openstack network log show {{params}} {log_name}'
    kwargs['format'] = 'json'
    return _client.execute(cmd, *args, **kwargs)


def network_log_list(*args, **kwargs):
    cmd = 'openstack network log list {params}'
    kwargs['format'] = 'json'
    return _client.execute(cmd, *args, **kwargs)


def network_log_delete(log_names, *args, **kwargs):
    cmd = f'openstack network log delete {{params}} {" ".join(log_names)}'
    return _client.execute(cmd, *args, **kwargs)



from __future__ import absolute_import

import functools
import os
import time
import typing

from oslo_log import log
import pandas

import tobiko
from tobiko import config
from tobiko.openstack import neutron
from tobiko.openstack import topology
from tobiko import rhosp as rhosp_topology
from tobiko.rhosp import containers as rhosp_containers


CONF = config.CONF
LOG = log.getLogger(__name__)


class ContainerRuntimeFixture(tobiko.SharedFixture):

    runtime: typing.Optional[rhosp_containers.ContainerRuntime] = None

    def setup_fixture(self):
        self.runtime = self.get_runtime()

    def cleanup_fixture(self):
        self.runtime = None

    @staticmethod
    def get_runtime() -> typing.Optional[rhosp_containers.ContainerRuntime]:
        """return handle to the container runtime"""
        return rhosp_containers.PODMAN_RUNTIME


def get_container_runtime() -> rhosp_containers.ContainerRuntime:
    runtime = tobiko.setup_fixture(ContainerRuntimeFixture).runtime
    return runtime


def get_container_runtime_name() -> str:
    return get_container_runtime().runtime_name


def is_docker() -> bool:
    return False


def is_podman() -> bool:
    return True


def has_container_runtime() -> bool:
    return True


@functools.lru_cache()
def list_node_containers(ssh_client):
    """returns a list of containers and their run state"""
    return get_container_runtime().list_containers(ssh_client=ssh_client)


def get_container_client(ssh_client=None):
    """returns a list of containers and their run state"""
    return get_container_runtime().get_client(ssh_client=ssh_client)


def list_containers_df(group=None):
    actual_containers_list = list_containers(group)
    return pandas.DataFrame(
        get_container_states_list(actual_containers_list),
        columns=['container_host', 'container_name', 'container_state'])


def list_containers(group=None):
    """get list of containers in running state
    from specified node group
    returns : a list of overcloud_node's running containers"""

    # moved here from topology
    # reason : Workaround for :
    # AttributeError: module 'tobiko.openstack.topology' has no
    # attribute 'container_runtime'

    if group is None:
        group = 'compute'
    containers_list = tobiko.Selection()
    openstack_nodes = topology.list_openstack_nodes(group=group)

    for node in openstack_nodes:
        LOG.debug(f"List containers for node {node.name}")
        node_containers_list = list_node_containers(ssh_client=node.ssh_client)
        containers_list.extend(node_containers_list)
    return containers_list


def assert_containers_running(group, expected_containers, full_name=True,
                              bool_check=False, nodenames=None):

    """assert that all containers specified in the list are running
    on the specified openstack group(controller or compute etc..)
    if bool_check is True then return only True or false without failing"""

    if is_docker():
        LOG.info('not checking common containers since we are on docker')
        return

    failures = []

    openstack_nodes = topology.list_openstack_nodes(group=group,
                                                    hostnames=nodenames)
    for node in openstack_nodes:
        node_containers = list_node_containers(ssh_client=node.ssh_client)
        containers_list_df = pandas.DataFrame(
            get_container_states_list(node_containers),
            columns=['container_host', 'container_name', 'container_state'])
        # check that the containers are present
        LOG.info('node: {} containers list : {}'.format(
            node.name, containers_list_df.to_string(index=False)))
        for container in expected_containers:
            # get container attrs dataframe
            if full_name:
                container_attrs = containers_list_df.query(
                    'container_name == "{}"'.format(container))
            else:
                container_attrs = containers_list_df[
                    containers_list_df['container_name'].
                    str.contains(container)]
            # check if the container exists
            LOG.info('checking container: {}'.format(container))
            if container_attrs.empty:
                failures.append(
                    'expected container {} not found on node {} ! : \n\n'.
                    format(container, node.name))
            # if container exists, check it is running
            else:
                # only one running container is expected
                container_running_attrs = container_attrs.query(
                    'container_state=="running"')
                if container_running_attrs.empty:
                    failures.append(
                        'expected container {} is not running on node {} , '
                        'its state is {}! : \n\n'.format(
                            container, node.name,
                            container_attrs.container_state.values.item()))
                elif len(container_running_attrs) > 1:
                    failures.append(
                        'only one running container {} was expected on '
                        'node {}, but got {}! : \n\n'.format(
                            container, node.name,
                            len(container_running_attrs)))

    if not bool_check and failures:
        tobiko.fail(
            'container states mismatched:\n{}'.format('\n'.join(failures)),
            rhosp_containers.ContainerMismatchException)

    elif bool_check and failures:
        return False

    else:
        LOG.info('All specified containers are in running state! ')
        return True


def assert_ovn_containers_running():
    if not neutron.has_ovn():
        LOG.info("Networking OVN not configured")
        return
    ovn_containers = ['ovn_metadata_agent',
                      'ovn_controller']
    potential_groups = ['edpm-compute', 'edpm-networker']
    groups = [group for group in potential_groups if
              group in topology.get_openstack_topology().groups]
    for group in groups:
        assert_containers_running(group, ovn_containers, full_name=False)
    LOG.info("Networking OVN containers verified in running state")


def get_container_states_list(containers_list,
                              include_container_objects=False):
    container_states_list = tobiko.Selection()
    container_states_list.extend([comparable_container_keys(
        container, include_container_objects=include_container_objects) for
                                  container in containers_list])
    return container_states_list


def save_containers_state_to_file(expected_containers_list,):
    expected_containers_list_df = pandas.DataFrame(
        get_container_states_list(expected_containers_list),
        columns=['container_host', 'container_name', 'container_state'])
    expected_containers_list_df.to_csv(
        rhosp_containers.expected_containers_file)
    return rhosp_containers.expected_containers_file


def comparable_container_keys(container, include_container_objects=False):
    """returns the tuple : 'container_host','container_name',
    'container_state, container object if specified'
     """
    host_or_ip = container.client.base_url.netloc.rsplit('_')[1]
    nodes_matching = topology.list_openstack_nodes(hostnames=[host_or_ip])
    nodename = (nodes_matching[0].name
                if nodes_matching
                else rhosp_topology.ip_to_hostname(host_or_ip))

    # Differenciate between podman_ver3 with podman-py from earlier api
    if include_container_objects:
        return (nodename,
                container.attrs['Names'][0], container.attrs['State'],
                container)
    else:
        return (nodename,
                container.attrs['Names'][0],
                container.attrs['State'])


@functools.lru_cache()
def list_containers_objects_df():
    containers_list = list_containers()
    containers_objects_list_df = pandas.DataFrame(
        get_container_states_list(
            containers_list, include_container_objects=True),
        columns=['container_host', 'container_name',
                 'container_state', 'container_object'])
    return containers_objects_list_df


def get_edpm_container(container_name=None, container_host=None,
                       partial_container_name=None):
    """gets an container object by name on specified host
    container"""
    con_obj_df = list_containers_objects_df()
    if partial_container_name and container_host:
        con_obj_df = con_obj_df[con_obj_df['container_name'].str.contains(
            partial_container_name)]
        contaniner_obj = con_obj_df.query(
            'container_host == "{container_host}"'.format(
                container_host=container_host))['container_object']
    elif container_host:
        contaniner_obj = con_obj_df.query(
            'container_name == "{container_name}"'
            ' and container_host == "{container_host}"'.
            format(container_host=container_host,
                   container_name=container_name)).container_object
    else:
        contaniner_obj = con_obj_df.query(
            'container_name == "{container_name}"'.
            format(container_name=container_name)).container_object
    if not contaniner_obj.empty:
        return contaniner_obj.values[0]
    else:
        tobiko.fail('container {} not found!'.format(container_name))


def action_on_container(action: str,
                        container_name=None,
                        container_host=None,
                        partial_container_name=None):
    """take a container and perform an action on it
    actions are as defined in : podman/libs/containers.py:14/164"""

    LOG.debug(f"Executing '{action}' action on container "
              f"'{container_name}@{container_host}'...")
    container = get_edpm_container(
        container_name=container_name,
        container_host=container_host,
        partial_container_name=partial_container_name)

    container_class: typing.Type = type(container)
    # we get the specified action as function from podman lib
    action_method: typing.Optional[typing.Callable] = getattr(
        container_class, action, None)
    if action_method is None:
        raise TypeError(f"Unsupported container action for class :"
                        f" {container_class}")
    if not callable(action_method):
        raise TypeError(
            f"Attribute '{container_class.__qualname__}.{action}' value "
            f" is not a method: {action_method!r}")
    LOG.debug(f"Calling '{action_method}' action on container "
              f"'{container}'")
    return action_method(container)


def assert_equal_containers_state(expected_containers_list=None,
                                  timeout=120, interval=2,
                                  recreate_expected=False):

    """compare all edpm container states with using two lists:
    one is current , the other some past list
    first time this method runs it creates a file holding overcloud
    containers' states: ~/expected_containers_list_df.csv'
    second time it creates a current containers states list and
    compares them, they must be identical"""

    expected_containers_list_df = []
    # if we have a file or an explicit variable use that , otherwise  create
    # and return
    if recreate_expected or (
            not expected_containers_list and
            not os.path.exists(rhosp_containers.expected_containers_file)):
        save_containers_state_to_file(list_containers())
        return

    elif expected_containers_list:
        expected_containers_list_df = pandas.DataFrame(
            get_container_states_list(expected_containers_list),
            columns=['container_host', 'container_name', 'container_state'])

    elif os.path.exists(rhosp_containers.expected_containers_file):
        expected_containers_list_df = pandas.read_csv(
            rhosp_containers.expected_containers_file)

    failures = []
    start = time.time()
    error_info = 'Output explanation: left_only is the original state, ' \
                 'right_only is the new state'

    while time.time() - start < timeout:

        failures = []
        actual_containers_list_df = list_containers_df()

        LOG.info('expected_containers_list_df: {} '.format(
            expected_containers_list_df.to_string(index=False)))
        LOG.info('actual_containers_list_df: {} '.format(
            actual_containers_list_df.to_string(index=False)))

        # execute a `dataframe` diff between the expected and actual containers
        expected_containers_state_changed = \
            rhosp_containers.dataframe_difference(
                    expected_containers_list_df,
                    actual_containers_list_df)
        # check for changed state containerstopology
        if not expected_containers_state_changed.empty:
            failures.append('expected containers changed state ! : '
                            '\n\n{}\n{}'.format(
                             expected_containers_state_changed.
                             to_string(index=False), error_info))
            LOG.info('container states mismatched:\n{}\n'.format(failures))
            time.sleep(interval)
            # clear cache to obtain new data
            list_node_containers.cache_clear()
        else:
            LOG.info("assert_equal_containers_state :"
                     " OK, all containers are on the same state")
            return
    if failures:
        tobiko.fail('container states mismatched:\n{!s}', '\n'.join(
            failures))



from __future__ import absolute_import

import typing

from oslo_log import log
import pandas

import tobiko
from tobiko.tripleo import overcloud
from tobiko.shell import sh
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


class OvercloudServiceException(tobiko.TobikoException):
    message = "not all overcloud nodes services are in active state"


def get_overcloud_node_services_table(ssh_client: ssh.SSHClientType):
    """
    get services table from overcloud node

       returns :
auditd.service|loaded|active|running|SecurityAuditingService
auth-rpcgss-module.service|loaded|inactivedead|KernelModulesupportingRPCSEC_GSS
blk-availability.service|loaded|active|exited|Availabilityofblockdevices
brandbot.service|loaded|inactivedead|FlexibleBrandingService
certmonger.service|loaded|active|running|CertificatemonitoringandPKIenrollment
cinder-lvm-losetup.service|loaded|inactivedead|CinderLVMlosetup
cloud-config.service|loaded|active|exited|Applythesettingsspecifiedincloud-con
cloud-final.service|loaded|active|exited|Executeclouduser/finalscripts
cloud-init-local.service|loaded|active|exited|Initialcloud-initjob(pre-network)
cloud-init.service|loaded|active|exited|Initialcloud-initjob(metadataservicecr)

    :return: dataframe of overcloud node services
    """
    units = sh.list_systemd_units(all=True,
                                  ssh_client=ssh_client).without_attributes(
        load='not-found')

    data: typing.Dict[str, list] = {'UNIT': [],
                                    'loaded_state': [],
                                    'active_state': [],
                                    'low_level_state': [],
                                    'UNIT_DESCRIPTION': []}
    for unit in units:
        data['UNIT'].append(unit.unit)
        data['loaded_state'].append(unit.load)
        data['active_state'].append(unit.active)
        data['low_level_state'].append(unit.sub)
        data['UNIT_DESCRIPTION'].append(unit.description)
    table = pandas.DataFrame.from_dict(data)
    table.replace(to_replace=' ', value="", regex=True, inplace=True)
    table['overcloud_node'] = sh.get_hostname(ssh_client=ssh_client)

    LOG.debug("Got overcloud nodes services status :\n%s", table)
    return table


def get_overcloud_nodes_running_service(service):
    """
    Check what nodes are running the specified service or unit
    process: exact str of a process name as seen in systemctl -a
    :return: list of overcloud nodes
    """
    oc_procs_df = overcloud.get_overcloud_nodes_dataframe(
                                            get_overcloud_node_services_table)
    # remove the ".service" suffix
    oc_procs_df = oc_procs_df.replace(to_replace={'UNIT': '.service'},
                                      value='',
                                      regex=True)
    oc_nodes_running_service = oc_procs_df.query('UNIT=="{}"'.format(service))[
                                                'overcloud_node'].unique()
    return oc_nodes_running_service.tolist()


def check_if_process_running_on_overcloud(process):
    """
    Check what nodes are running the specifies
    process: exact str of a process name as seen in ps -axw -o "%c"
    :return: list of overcloud nodes
    """
    oc_services_df = overcloud.get_overcloud_nodes_dataframe(
                                            get_overcloud_node_services_table)
    if not oc_services_df.query('UNIT=="{}"'.format(process)).empty:
        return True
    else:
        return False


class OvercloudServicesStatus(tobiko.SharedFixture):
    """
    class to handle services checks,
    checks that all of these are running in the overcloud:
    'corosync.service','iptables.service','network.service','ntpd.service',
    'pacemaker.service','rpcbind.service','sshd.service'

    """

    SERVICES_TO_CHECK: typing.List[str] = [
        'corosync.service',
        'iptables.service',
        'network.service',
        # Not found on OSP 16
        # 'ntpd.service',
        'pacemaker.service',
        'rpcbind.service',
        'sshd.service']

    def __init__(self,
                 services_to_check: typing.List[str] = None):
        super().__init__()
        if services_to_check is None:
            services_to_check = self.SERVICES_TO_CHECK
        self.services_to_check = services_to_check

    oc_services_df: typing.Any

    def setup_fixture(self):
        self.oc_services_df = overcloud.get_overcloud_nodes_dataframe(
            get_overcloud_node_services_table)

    @property
    def basic_overcloud_services_running(self):
        """
        Checks that the oc_services dataframe has all of the list services
        running
        :return: Bool
        """
        tobiko.setup_fixture(self)
        for service_name in self.services_to_check:
            if not self.oc_services_df.query('UNIT=="{}"'.format(
                    service_name)).empty:
                LOG.info("overcloud processes status checks: process {} is  "
                         "in running state".format(service_name))
                continue
            else:
                LOG.info("Failure : overcloud processes status checks: "
                         "process {} is not running ".format(service_name))
                raise OvercloudServiceException()
        return True



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import unittest

from packaging import version

import tobiko


class TestVersion(unittest.TestCase):

    def test_parse_version(self):
        self.assertEqual(version.parse('1.0.0'),
                         tobiko.parse_version('1'))
        self.assertEqual(version.parse('1.2.0'),
                         tobiko.parse_version('1.2'))
        self.assertEqual(version.parse('1.2.3'),
                         tobiko.parse_version('1.2.3'))
        self.assertEqual(version.parse('3.2.1'),
                         tobiko.parse_version(' a b c 10 3.2.1 25.2.1'))

    def test_parse_version_with_invalid(self):
        with self.assertRaises(tobiko.InvalidVersion):
            tobiko.parse_version('abc')

    def test_get_version_with_srt(self):
        self.assertEqual(version.parse('3.2.1'),
                         tobiko.get_version(' a b c 10 3.2.1 25.2.1'))

    def test_get_version_with_version(self):
        reference = version.parse('3.2.1')
        self.assertIs(reference,
                      tobiko.get_version(reference))

    def test_match_version_with_str(self):
        self.assertTrue(tobiko.match_version('1.2.3'))

    def test_match_version_with_version(self):
        reference = version.parse('3.2.1')
        self.assertTrue(tobiko.match_version(reference))

    def test_match_version_min_version(self):
        self.assertTrue(tobiko.match_version('1.0.0',
                                             min_version='0.9.9'))
        self.assertTrue(tobiko.match_version('1.0.0',
                                             min_version='1'))
        self.assertTrue(tobiko.match_version('1.0.0',
                                             min_version='1.0'))
        self.assertTrue(tobiko.match_version('1.0.0',
                                             min_version='1.0.0'))
        self.assertFalse(tobiko.match_version('1.0.0',
                                              min_version='1.0.1'))

    def test_match_version_max_version(self):
        self.assertFalse(tobiko.match_version('1.0.0',
                                              max_version='0.9.9'))
        self.assertFalse(tobiko.match_version('1.0.0',
                                              max_version='1.0.0'))
        self.assertTrue(tobiko.match_version('1.0.0',
                                             max_version='1.0.1'))



"""Model for accessing details of Pods from podman service."""

from __future__ import absolute_import

import collections
import json
import signal
import time

from . import ConfigDict, FoldedString, fold_keys


class Pod(collections.UserDict):
    """Model for a Pod."""

    def __init__(self, client, ident, data):
        """Construct Pod model."""
        super().__init__(data)

        self._ident = ident
        self._client = client

        with client() as podman:
            self._refresh(podman)

    def _refresh(self, podman):
        pod = podman.GetPod(self._ident)
        super().update(pod['pod'])

        for k, v in self.data.items():
            setattr(self, k, v)
        return self

    def inspect(self):
        """Retrieve details about pod."""
        with self._client() as podman:
            results = podman.InspectPod(self._ident)
        obj = json.loads(results['pod'], object_hook=fold_keys())
        obj['id'] = obj['config']['id']
        return collections.namedtuple('PodInspect', obj.keys())(**obj)

    def kill(self, signal_=signal.SIGTERM, wait=25):
        """Send signal to all containers in pod.

        default signal is signal.SIGTERM.
        wait n of seconds, 0 waits forever.
        """
        with self._client() as podman:
            podman.KillPod(self._ident, signal_)
            timeout = time.time() + wait
            while True:
                # pylint: disable=maybe-no-member
                self._refresh(podman)
                running = FoldedString(self.status)
                if running != 'running':
                    break

                if wait and timeout < time.time():
                    raise TimeoutError()

                time.sleep(0.5)
        return self

    def pause(self):
        """Pause all containers in the pod."""
        with self._client() as podman:
            podman.PausePod(self._ident)
            return self._refresh(podman)

    def refresh(self):
        """Refresh status fields for this pod."""
        with self._client() as podman:
            return self._refresh(podman)

    def remove(self, force=False):
        """Remove pod and its containers returning pod ident.

        force=True, stop any running container.
        """
        with self._client() as podman:
            results = podman.RemovePod(self._ident, force)
        return results['pod']

    def restart(self):
        """Restart all containers in the pod."""
        with self._client() as podman:
            podman.RestartPod(self._ident)
            return self._refresh(podman)

    def stats(self):
        """Stats on all containers in the pod."""
        with self._client() as podman:
            results = podman.GetPodStats(self._ident)
        for obj in results['containers']:
            yield collections.namedtuple('ContainerStats', obj.keys())(**obj)

    def start(self):
        """Start all containers in the pod."""
        with self._client() as podman:
            podman.StartPod(self._ident)
            return self._refresh(podman)

    def stop(self):
        """Stop all containers in the pod."""
        with self._client() as podman:
            podman.StopPod(self._ident)
            return self._refresh(podman)

    def top(self):
        """Display stats for all containers."""
        with self._client() as podman:
            results = podman.TopPod(self._ident)
        return results['pod']

    def unpause(self):
        """Unpause all containers in the pod."""
        with self._client() as podman:
            podman.UnpausePod(self._ident)
            return self._refresh(podman)


class Pods():
    """Model for accessing pods."""

    def __init__(self, client):
        """Construct pod model."""
        self._client = client

    def create(self,
               ident=None,
               cgroupparent=None,
               labels=None,
               share=None,
               infra=False):
        """Create a new empty pod."""
        config = ConfigDict(
            name=ident,
            cgroupParent=cgroupparent,
            labels=labels,
            share=share,
            infra=infra,
        )

        with self._client() as podman:
            result = podman.CreatePod(config)
            details = podman.GetPod(result['pod'])
        return Pod(self._client, result['pod'], details['pod'])

    def get(self, ident):
        """Get Pod from ident."""
        with self._client() as podman:
            result = podman.GetPod(ident)
        return Pod(self._client, result['pod']['id'], result['pod'])

    def list(self):
        """List all pods."""
        with self._client() as podman:
            results = podman.ListPods()
        for pod in results['pods']:
            yield Pod(self._client, pod['id'], pod)



# Copyright (c) 2024 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import contextlib
import functools
import os

from oslo_concurrency import lockutils
from oslo_log import log
from oslo_utils import reflection
from oslo_utils import timeutils


LOG = log.getLogger(__name__)


def interworker_synched(name):
    """Re-definition of oslo_concurrency.lockutils.synchronized.

    Tobiko needs to re-difine this decorator in order to avoid intra-
    process/worker locks. This is because tobiko is executed in multiple
    processes (using pytest), but each process only runs one thread.

    The intra-process lock should not be applied in tobiko because some of the
    locked methods could be called recurrently.
    Example:
    The creation (setup_fixture) of CirrosPeerServerStackFixture depends on the
    creation of CirrosServerStackFixture, which is also its parent class.
    With intra-process locks, the creation of CirrosServerStackFixture could
    not be started (would be locked by the creation of
    CirrosPeerServerStackFixture).
    """

    def wrap(f):

        @functools.wraps(f)
        def inner(*args, **kwargs):
            t1 = timeutils.now()
            t2 = None
            gotten = True
            f_name = reflection.get_callable_name(f)
            try:
                with lock(name):
                    t2 = timeutils.now()
                    LOG.debug('Lock "%(name)s" acquired by "%(function)s" :: '
                              'waited %(wait_secs)0.3fs',
                              {'name': name,
                               'function': f_name,
                               'wait_secs': (t2 - t1)})
                    return f(*args, **kwargs)
            except lockutils.AcquireLockFailedException:
                gotten = False
            finally:
                t3 = timeutils.now()
                if t2 is None:
                    held_secs = "N/A"
                else:
                    held_secs = "%0.3fs" % (t3 - t2)
                LOG.debug('Lock "%(name)s" "%(gotten)s" by "%(function)s" ::'
                          ' held %(held_secs)s',
                          {'name': name,
                           'gotten': 'released' if gotten else 'unacquired',
                           'function': f_name,
                           'held_secs': held_secs})
        return inner

    return wrap


@contextlib.contextmanager
def lock(name):
    """Re-definition of oslo_concurrency.lockutils.lock that does not apply
    intra-worker locks. Only inter-worker locks are applied.
    """
    from tobiko import config
    lock_path = os.path.expanduser(config.CONF.tobiko.common.lock_dir)

    LOG.debug('Acquired lock "%(lock)s"', {'lock': name})

    try:
        ext_lock = lockutils.external_lock(name,
                                           lock_file_prefix='tobiko',
                                           lock_path=lock_path)
        gotten = ext_lock.acquire(delay=0.01, blocking=True)
        if not gotten:
            raise lockutils.AcquireLockFailedException(name)
        LOG.debug('Acquired external semaphore "%(lock)s"',
                  {'lock': name})
        try:
            yield ext_lock
        finally:
            ext_lock.release()
    finally:
        LOG.debug('Releasing lock "%(lock)s"', {'lock': name})



# Copyright 2023 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import itertools

from oslo_config import cfg

GROUP_NAME = 'manila'
OPTIONS = [
    cfg.StrOpt('share_protocol',
               default='nfs',
               help="Share protocol"),
    cfg.IntOpt('size',
               default=1,
               help="Default size in GB for shares created by share tests."),
    cfg.StrOpt('default_share_type_name',
               default='default',
               help="Default share type's name"),
    cfg.BoolOpt('spec_driver_handles_share_servers',
                default=False,
                help="Specifies whether the driver handles the share servers "
                     "or not"),
]


def register_tobiko_options(conf):
    conf.register_opts(group=cfg.OptGroup(GROUP_NAME), opts=OPTIONS)


def list_options():
    return [(GROUP_NAME, itertools.chain(OPTIONS))]



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import re
import typing

from oslo_log import log

import tobiko
from tobiko.shell.sh import _command
from tobiko.shell.sh import _execute
from tobiko.shell import ssh

# pylint: disable=redefined-builtin

LOG = log.getLogger(__name__)


class SystemdUnit(typing.NamedTuple):
    unit: str
    load: str
    active: str
    sub: str
    description: str
    data: typing.Dict[str, typing.Any]


IS_WORD_PATTERN = re.compile(r'\S+')


SystemdUnitType = typing.Union[str, SystemdUnit]


def get_systemd_unit_names(*units: SystemdUnitType) -> tobiko.Selection[str]:
    return tobiko.select(get_systemd_unit_name(unit)
                         for unit in units)


def get_systemd_unit_name(unit: SystemdUnitType) -> str:
    if isinstance(unit, str):
        return unit
    else:
        return tobiko.check_valid_type(unit, SystemdUnit).unit


def systemctl_command(command: str,
                      *units: SystemdUnitType,
                      all: bool = None,
                      no_pager: bool = None,
                      plain: bool = None,
                      state: str = None,
                      type: str = None) \
        -> _command.ShellCommand:
    command_line = _command.shell_command('systemctl') + command
    if all:
        command_line += '-a'
    if no_pager:
        command_line += '--no-pager'
    if plain:
        command_line += '--plain'
    if state is not None:
        command_line += f'"--state={state}"'
    if type is not None:
        command_line += f'-t "{type}"'
    if units:
        command_line += get_systemd_unit_names(*units)
    return command_line


def list_systemd_units(*units: SystemdUnitType,
                       all: bool = None,
                       state: str = None,
                       type: str = None,
                       ssh_client: ssh.SSHClientType = None,
                       sudo: bool = None) \
        -> tobiko.Selection[SystemdUnit]:
    command = systemctl_command('list-units', *units, all=all,
                                no_pager=True, plain=True, state=state,
                                type=type)
    output = _execute.execute(command,
                              ssh_client=ssh_client,
                              sudo=sudo).stdout
    result = tobiko.Selection[SystemdUnit]()
    if output.startswith('0 loaded units listed.'):
        return result

    table_lines = iter(output.splitlines())
    first_line = next(table_lines)
    search_pos = 0
    names: typing.List[str] = []
    starts: typing.List[int] = []
    while True:
        match = IS_WORD_PATTERN.search(first_line, search_pos)
        if match is None:
            break
        names.append(match.group(0).lower())
        starts.append(match.start())
        search_pos = match.end()
    ends: typing.List[int] = list(starts[1:]) + [-1]

    for line in table_lines:
        message = line.strip()
        if not message:
            # Legend is coming after the first empty line
            break

        if message == '0 loaded units listed.':
            raise SystemdUnitNotFound(command=command, output=output)

        data: typing.Dict[str, str] = {}
        for name, start, end in zip(names, starts, ends):
            data[name] = line[start:end].strip()
        result.append(SystemdUnit(unit=data.get('unit', ''),
                                  load=data.get('load', ''),
                                  active=data.get('active', ''),
                                  sub=data.get('sub', ''),
                                  description=data.get('description', ''),
                                  data=data))
    return result


def stop_systemd_units(*units: SystemdUnitType,
                       ssh_client: ssh.SSHClientType = None):
    command = systemctl_command('stop', *units)
    _execute.execute(command, ssh_client=ssh_client, sudo=True)


def start_systemd_units(*units: SystemdUnitType,
                        ssh_client: ssh.SSHClientType = None):
    command = systemctl_command('start', *units)
    _execute.execute(command, ssh_client=ssh_client, sudo=True)


def wait_for_active_systemd_units(*units: SystemdUnitType,
                                  state: str = None,
                                  type: str = None,
                                  timeout: tobiko.Seconds = None,
                                  interval: tobiko.Seconds = None,
                                  ssh_client: ssh.SSHClientType = None,
                                  sudo: bool = None) \
        -> tobiko.Selection[SystemdUnit]:
    return wait_for_systemd_units_state(match_unit_state(active='ACTIVE'),
                                        *units,
                                        state=state,
                                        type=type,
                                        timeout=timeout,
                                        interval=interval,
                                        ssh_client=ssh_client,
                                        sudo=sudo)


MATCH_ALL = re.compile(r'.*')
Pattern = type(MATCH_ALL)
PatternType = typing.Union[str, typing.Pattern[str]]


def compile_pattern(pattern: typing.Optional[PatternType]) \
        -> typing.Pattern[str]:
    if pattern is None:
        return MATCH_ALL
    elif isinstance(pattern, str):
        return re.compile(pattern, re.IGNORECASE)
    return tobiko.check_valid_type(pattern, Pattern)


class MatchUnitState(typing.NamedTuple):
    load: typing.Pattern[str] = MATCH_ALL
    active: typing.Pattern[str] = MATCH_ALL
    sub: typing.Pattern[str] = MATCH_ALL

    def match_unit(self, unit: SystemdUnit) -> bool:
        return bool(self.load.match(unit.load) and
                    self.active.match(unit.active) and
                    self.sub.match(unit.sub))

    def __call__(self, unit: SystemdUnit) -> bool:
        return self.match_unit(unit)

    def __repr__(self) -> str:
        details = []
        if self.load != MATCH_ALL:
            details.append(f"load={self.load.pattern!r}")
        if self.active != MATCH_ALL:
            details.append(f"load={self.active.pattern!r}")
        if self.sub != MATCH_ALL:
            details.append(f"load={self.sub.pattern!r}")
        details_text = ', '.join(details)
        return f'{type(self).__name__}({details_text})'


def match_unit_state(load: PatternType = None,
                     active: PatternType = None,
                     sub: PatternType = None) \
        -> typing.Callable[[SystemdUnit], bool]:
    return typing.cast(
        typing.Callable[[SystemdUnit], bool],
        MatchUnitState(load=compile_pattern(load),
                       active=compile_pattern(active),
                       sub=compile_pattern(sub)))


def wait_for_systemd_units_state(
        match_unit: typing.Callable[[SystemdUnit], bool],
        *units: SystemdUnitType,
        state: str = None,
        type: str = None,
        ssh_client: ssh.SSHClientType = None,
        sudo: bool = None,
        check: bool = True,
        timeout: tobiko.Seconds = None,
        interval: tobiko.Seconds = None) \
        -> tobiko.Selection[SystemdUnit]:
    missing_units = tobiko.select(units)
    all_units: typing.Dict[str, SystemdUnit] = collections.OrderedDict()
    for attempt in tobiko.retry(timeout=timeout,
                                interval=interval,
                                default_timeout=30.,
                                default_interval=5.):
        actual_units = list_systemd_units(*missing_units,
                                          all=True,
                                          state=state,
                                          type=type,
                                          ssh_client=ssh_client,
                                          sudo=sudo)

        all_units.update((unit.unit, unit) for unit in actual_units)
        missing_units = typing.cast(
            tobiko.Selection[SystemdUnitType],
            actual_units.select(match_unit, expect=False))
        if not missing_units:
            break

        LOG.info('Systemd unit(s) still on unexpected state:'
                 f' expected: ({match_unit})...:\n'
                 '  actual: \n'
                 '\n'.join(f'    - {u}' for u in missing_units))
        if attempt.is_last:
            break

    if check:
        if missing_units:
            raise UnexpectedSystemctlUnitState(matcher=match_unit,
                                               units=missing_units)
    return tobiko.Selection(all_units.values())


class SystemdUnitNotFound(tobiko.TobikoException):
    message = ("Systemd unit(s) not found\n"
               "  command: {command}\n"
               "  output: {output}")


class UnexpectedSystemctlUnitState(tobiko.TobikoException):
    message = ("Systemd unit(s) has unexpected state: \n"
               "  expected: {matcher}\n"
               "  actual: {units}")



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import multiprocessing
from multiprocessing import pool
import typing

import tobiko
from tobiko.run import _config


class WorkersPoolFixture(tobiko.SharedFixture):

    config = tobiko.required_fixture(_config.RunConfigFixture)

    pool: pool.Pool
    workers_count: int = 0

    def __init__(self, workers_count: int = None):
        super().__init__()
        if workers_count is not None:
            self.workers_count = workers_count

    def setup_fixture(self):
        workers_count = self.workers_count
        if not workers_count:
            workers_count = self.config.workers_count
        self.workers_count = workers_count or 0
        context = multiprocessing.get_context('spawn')
        self.pool = context.Pool(processes=workers_count or None)


def workers_pool() -> pool.Pool:
    return tobiko.setup_fixture(WorkersPoolFixture).pool


def call_async(func: typing.Callable,
               *args,
               **kwargs):
    return workers_pool().apply_async(func, args=args, kwds=kwargs)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import itertools

from oslo_config import cfg

DEFAULT_KEY_TYPE = 'ecdsa'
DEFAULT_KEY_FILE = f'~/.ssh/id_{DEFAULT_KEY_TYPE}'

GROUP_NAME = "nova"
OPTIONS = [
    cfg.StrOpt('key_file',
               default=DEFAULT_KEY_FILE,
               help="Default SSH key to login to server instances"),
    cfg.StrOpt('key_type',
               default=DEFAULT_KEY_TYPE,
               help="Default SSH key type to login to server instances"),
    cfg.FloatOpt('ubuntu_connection_timeout',
                 default=1500.,
                 help="Timeout (in seconds) for establishing connection "
                      "to ubuntu"),
    cfg.FloatOpt('ubuntu_is_reachable_timeout',
                 default=900.,
                 help="Timeout (in seconds) till ubuntu server is reachable"),
    cfg.FloatOpt('cloudinit_is_reachable_timeout',
                 default=600.,
                 help="Timeout (in seconds) till cloud-init based server is "
                      "reachable")
]


def register_tobiko_options(conf):
    conf.register_opts(group=cfg.OptGroup('nova'), opts=OPTIONS)


def list_options():
    return [(GROUP_NAME, itertools.chain(OPTIONS))]



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import functools

import requests
from urllib3 import poolmanager

from tobiko.http import _connection


def get_http_session(ssh_client=None):
    return setup_http_session(session=requests.Session(),
                              ssh_client=ssh_client)


def setup_http_session(session, ssh_client=None):
    for adapter in session.adapters.values():
        manager = adapter.poolmanager
        manager.pool_classes_by_scheme = pool_classes_by_scheme.copy()
        manager.key_fn_by_scheme = key_fn_by_scheme.copy()
        manager.connection_pool_kw['ssh_client'] = ssh_client
    return session


# pylint: disable=protected-access

# All known keyword arguments that could be provided to the pool manager, its
# pools, or the underlying connections. This is used to construct a pool key.
_key_fields = tuple(
    poolmanager._key_fields  # type: ignore
) + ('key_ssh_client',)


class PoolKey(collections.namedtuple("PoolKey", _key_fields)):  # type: ignore
    """The namedtuple class used to construct keys for the connection pool.

    All custom key schemes should include the fields in this key at a minimum.
    """


#: A dictionary that maps a scheme to a callable that creates a pool key.
#: This can be used to alter the way pool keys are constructed, if desired.
#: Each PoolManager makes a copy of this dictionary so they can be configured
#: globally here, or individually on the instance.
key_fn_by_scheme = {
    "http": functools.partial(
        poolmanager._default_key_normalizer,  # type: ignore
        PoolKey),
    "https": functools.partial(
        poolmanager._default_key_normalizer,  # type: ignore
        PoolKey),
}

# pylint: enable=protected-access


pool_classes_by_scheme = {"http": _connection.HTTPConnectionPool,
                          "https": _connection.HTTPSConnectionPool}



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os

import testtools

from tobiko.shell import sh


class WcTest(testtools.TestCase):

    def test_get_file_size(self):
        file_name = os.path.join(sh.make_temp_dir(),
                                 self.id())
        sh.execute(f'echo "{self.id()}" > "{file_name}"')
        file_size = sh.get_file_size(file_name)
        self.assertEqual(len(self.id()) + 1, file_size)

    def test_get_file_size_when_empty(self):
        file_name = os.path.join(sh.make_temp_dir(),
                                 self.id())
        sh.execute(f'touch "{file_name}"')
        file_size = sh.get_file_size(file_name)
        self.assertEqual(0, file_size)

    def test_get_file_size_when_not_found(self):
        file_name = os.path.join(sh.make_temp_dir(),
                                 self.id())
        ex = self.assertRaises(sh.ShellCommandFailed,
                               sh.get_file_size,
                               file_name)
        self.assertEqual(1, ex.exit_status)
        self.assertIn(
            ex.stderr.strip(),
            [f"wc: {file_name}: No such file or directory",
             f"wc: {file_name}: open: No such file or directory"])

    def test_assert_file_size(self):
        file_name = os.path.join(sh.make_temp_dir(),
                                 self.id())
        sh.execute(f'echo "{self.id()}" > "{file_name}"')
        sh.assert_file_size(len(self.id()) + 1, file_name)

    def test_assert_file_size_when_empty(self):
        file_name = os.path.join(sh.make_temp_dir(),
                                 self.id())
        sh.execute(f'touch "{file_name}"')
        sh.assert_file_size(0, file_name)

    def test_assert_file_size_when_not_found(self):
        file_name = os.path.join(sh.make_temp_dir(),
                                 self.id())
        ex = self.assertRaises(sh.ShellCommandFailed,
                               sh.assert_file_size,
                               1, file_name)
        self.assertEqual(1, ex.exit_status)
        self.assertIn(
            ex.stderr.strip(),
            [f"wc: {file_name}: No such file or directory",
             f"wc: {file_name}: open: No such file or directory"])

    def test_assert_file_size_when_mismatch(self):
        file_name = os.path.join(sh.make_temp_dir(),
                                 self.id())
        sh.execute(f'echo "{self.id()}" > "{file_name}"')
        ex = self.assertRaises(testtools.matchers.MismatchError,
                               sh.assert_file_size,
                               1, file_name)
        self.assertEqual(f'1 != {len(self.id()) + 1}', str(ex))



# Copyright (c) 2022 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import typing

from oslo_log import log

import tobiko
from tobiko.shell import sh
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


class AnsiblePlaybook(tobiko.SharedFixture):

    def __init__(self,
                 command: sh.ShellCommandType = 'ansible-playbook',
                 inventory_filenames: typing.Iterable[str] = None,
                 playbook: str = 'main',
                 playbook_dirname: str = None,
                 requirement_files: typing.Iterable[str] = None,
                 roles_path: typing.Iterable[str] = None,
                 ssh_client: ssh.SSHClientType = None,
                 verbosity: int = None,
                 work_dir: str = None):
        super(AnsiblePlaybook, self).__init__()
        self._command = sh.shell_command(command)
        if inventory_filenames is None:
            inventory_filenames = []
        self._inventory_filenames = list(inventory_filenames)
        self._playbook = playbook
        self._playbook_dirname = playbook_dirname
        if requirement_files is None:
            requirement_files = []
        self._requirement_files = list(requirement_files)
        self._roles_path = roles_path
        self._ssh_client = ssh_client
        self._verbosity = verbosity
        self._work_dir = work_dir
        self._work_files: typing.Dict[str, str] = {}

    def get_inventory_file(self, inventory_filename: str):
        pass

    @property
    def ssh_client(self) -> ssh.SSHClientType:
        return self._ssh_client

    _connection: typing.Optional[sh.ShellConnection] = None

    @property
    def connection(self) -> sh.ShellConnection:
        if self._connection is None:
            self._connection = sh.shell_connection(self.ssh_client)
        return self._connection

    @property
    def work_dir(self) -> str:
        if self._work_dir is None:
            self._work_dir = self.connection.make_temp_dir(auto_clean=False)
            self._work_files = {}
        return self._work_dir

    @property
    def playbook_dirname(self) -> str:
        return tobiko.check_valid_type(self._playbook_dirname, str)

    @property
    def roles_path(self) -> typing.List[str]:
        roles_path = self._roles_path
        if roles_path is None:
            if roles_path is None:
                roles_path = []
            else:
                roles_path = list(roles_path)
            playbook_dirname = self._playbook_dirname
            if playbook_dirname is not None:
                playbook_roles_dir = os.path.join(playbook_dirname, 'roles')
                roles_path = ([playbook_roles_dir] +
                              roles_path +
                              [playbook_dirname])
            self._roles_path = roles_path
        return list(roles_path)

    @property
    def verbosity(self) -> typing.Optional[int]:
        if self._verbosity is None:
            self._verbosity = tobiko.tobiko_config().ansible.verbosity
        return self._verbosity

    def _ensure_inventory_files(self, *inventory_filenames: str) \
            -> typing.List[str]:
        filenames = list(inventory_filenames)
        filenames.extend(self._inventory_filenames)
        filenames.extend(tobiko.tobiko_config().ansible.inventory_files)
        existing_filenames = []
        for filename in sorted(filenames):
            filename = tobiko.tobiko_config_path(filename)
            if (filename not in existing_filenames and
                    os.path.isfile(filename)):
                existing_filenames.append(filename)
        if existing_filenames:
            self._ensure_work_files(*existing_filenames, sub_dir='inventory')
            return [os.path.join(self.work_dir, 'inventory')]
        dump_filenames = '  \n'.join(filenames)
        LOG.warning("Any Ansible inventory file(s) found:\n"
                    f"  {dump_filenames}\n")
        return []

    def _get_playbook_filename(self,
                               basename: str = None,
                               dirname: str = None) -> str:
        if basename is None:
            basename = self._playbook
        if dirname is None:
            dirname = self.playbook_dirname
        return os.path.join(dirname, basename)

    def _ensure_roles(self, roles: typing.Iterable[str],
                      dirname: str = None,
                      roles_path: typing.Iterable[str] = None) \
            -> typing.List[str]:
        role_dirs = []
        for role in roles:
            if roles_path is None:
                roles_path = self.roles_path
            else:
                roles_path = list(roles_path)
            if dirname is not None:
                dirname = os.path.realpath(dirname)
                roles_path = ([os.path.join(dirname, 'roles')] +
                              roles_path +
                              [dirname])
            for roles_dir in roles_path:
                role_dir = os.path.join(roles_dir, role)
                if os.path.isdir(role_dir):
                    role_dirs.append(role_dir)
                    break
            else:
                raise ValueError(
                    f'Role {role} not found in directories {self.roles_path}')
        return self._ensure_work_files(*role_dirs, sub_dir='roles')

    def _ensure_playbook_files(self,
                               playbook_files: typing.Iterable[str],
                               sub_dir: str = None,
                               dirname: str = None) -> typing.List[str]:
        work_filenames = []
        for playbook_file in playbook_files:
            filename = self._get_playbook_filename(basename=playbook_file,
                                                   dirname=dirname)
            if sub_dir is None and dirname is not None:
                if filename.startswith(dirname):
                    sub_dir = os.path.relpath(os.path.dirname(filename),
                                              dirname)

            work_filename = self._ensure_work_file(filename=filename,
                                                   sub_dir=sub_dir)
            work_filenames.append(work_filename)
        return work_filenames

    def _ensure_work_file(self, filename: str, sub_dir: str = None) -> str:
        filename = os.path.realpath(filename)
        work_filename = self._work_files.get(filename)
        if work_filename is None:
            if sub_dir is not None:
                work_filename = os.path.join(
                    self.work_dir, sub_dir, os.path.basename(filename))
            else:
                work_filename = os.path.join(
                    self.work_dir, os.path.basename(filename))
            if sub_dir is not None:
                self.connection.make_dirs(os.path.dirname(work_filename))
            self.connection.put_file(filename, work_filename)
            self._work_files[filename] = work_filename
        return work_filename

    def _ensure_work_files(self, *filenames: str, sub_dir: str = None) \
            -> typing.List[str]:
        missing_filenames = set()
        work_filenames = set()
        for filename in filenames:
            filename = os.path.realpath(filename)
            work_filename = self._work_files.get(filename)
            if work_filename is None:
                missing_filenames.add(filename)
            else:
                work_filenames.add(work_filename)
        if missing_filenames:
            if sub_dir is None:
                work_dir = self.work_dir
            else:
                work_dir = os.path.join(self.work_dir, sub_dir)
            self.connection.put_files(*sorted(missing_filenames),
                                      remote_dir=work_dir)
            for filename in filenames:
                work_filename = os.path.join(
                    work_dir, os.path.basename(filename))
                self._work_files[filename] = work_filename
                work_filenames.add(work_filename)
        return sorted(work_filenames)

    def cleanup_fixture(self):
        self._connection = None
        self._work_files = None
        if self._work_dir is not None:
            self.connection.remove_files(self._work_dir)
            self._work_dir = None

    def _get_command(self,
                     command: sh.ShellCommandType = None,
                     playbook: str = None,
                     playbook_dirname: str = None,
                     playbook_filename: str = None,
                     inventory_filenames: typing.Iterable[str] = None,
                     playbook_files: typing.Iterable[str] = None,
                     requirements_files: typing.Iterable[str] = None,
                     roles: typing.Iterable[str] = None,
                     roles_path: typing.Iterable[str] = None,
                     verbosity: int = None) -> \
            sh.ShellCommand:
        # ensure command
        if command is None:
            command = self._command
        assert isinstance(command, sh.ShellCommand)

        if verbosity is None:
            verbosity = self.verbosity
        if verbosity is not None and verbosity > 0:
            command += '-' + ('v' * verbosity)

        # ensure inventory
        if inventory_filenames is None:
            inventory_filenames = []

        for inventory_work_file in self._ensure_inventory_files(*list(
                inventory_filenames)):
            command += ['-i', inventory_work_file]

        # ensure playbook file
        if playbook_filename is None:
            playbook_filename = self._get_playbook_filename(
                basename=playbook, dirname=playbook_dirname)
        else:
            playbook_filename = os.path.realpath(playbook_filename)
        playbook_dirname = os.path.dirname(playbook_filename)
        command += [self._ensure_work_file(playbook_filename)]

        if playbook_files is not None:
            self._ensure_playbook_files(playbook_files=playbook_files,
                                        dirname=playbook_dirname)

        if roles is not None:
            self._ensure_roles(roles=roles,
                               dirname=playbook_dirname,
                               roles_path=roles_path)

        self._ensure_collections(requirements_files=requirements_files,
                                 dirname=playbook_dirname)
        return command

    def _ensure_collections(self,
                            requirements_files: typing.Iterable[str] = None,
                            dirname: str = None):
        work_files = self._ensure_requirements_files(
            requirements_files=requirements_files,
            dirname=dirname)
        if work_files:
            collections_dirname = os.path.join(self.work_dir, 'collections')
            command = sh.shell_command('ansible-galaxy collection install')
            command += ['-p', collections_dirname]
            for work_file in work_files:
                command += ['-r', work_file]
            self.connection.execute(command=command)

    def _ensure_requirements_files(
            self,
            requirements_files: typing.Iterable[str] = None,
            dirname: str = None) \
            -> typing.List[str]:
        if requirements_files is None:
            requirements_files = []
        else:
            requirements_files = list(requirements_files)
        for filename in self._requirement_files:
            if filename not in requirements_files:
                requirements_files.append(filename)
        if requirements_files:
            return self._ensure_playbook_files(
                playbook_files=requirements_files,
                sub_dir='requirements',
                dirname=dirname)
        else:
            return []

    def run_playbook(self,
                     command: sh.ShellCommand = None,
                     playbook: str = None,
                     playbook_dirname: str = None,
                     playbook_filename: str = None,
                     inventory_filenames: typing.Iterable[str] = None,
                     playbook_files: typing.Iterable[str] = None,
                     requirements_files: typing.Iterable[str] = None,
                     roles: typing.Iterable[str] = None,
                     roles_path: typing.Iterable[str] = None,
                     verbosity: int = None):
        tobiko.setup_fixture(self)
        command = self._get_command(command=command,
                                    playbook=playbook,
                                    playbook_dirname=playbook_dirname,
                                    playbook_filename=playbook_filename,
                                    inventory_filenames=inventory_filenames,
                                    playbook_files=playbook_files,
                                    requirements_files=requirements_files,
                                    roles=roles,
                                    roles_path=roles_path,
                                    verbosity=verbosity)
        return self.connection.execute(command, current_dir=self.work_dir)


def local_ansible_playbook() -> 'AnsiblePlaybook':
    return tobiko.get_fixture(AnsiblePlaybook)


def ansible_playbook(ssh_client: ssh.SSHClientType = None,
                     manager: 'AnsiblePlaybookManager' = None) -> \
        'AnsiblePlaybook':
    return ansible_playbook_manager(manager).get_ansible_playbook(
        ssh_client=ssh_client)


def register_ansible_playbook(playbook: 'AnsiblePlaybook',
                              manager: 'AnsiblePlaybookManager' = None) -> \
        None:
    tobiko.check_valid_type(playbook, AnsiblePlaybook)
    ansible_playbook_manager(manager).register_ansible_playbook(playbook)


def ansible_playbook_manager(manager: 'AnsiblePlaybookManager' = None) -> \
        'AnsiblePlaybookManager':
    if manager is None:
        return tobiko.setup_fixture(AnsiblePlaybookManager)
    else:
        tobiko.check_valid_type(manager, AnsiblePlaybookManager)
        return manager


AnsiblePlaybookKey = typing.Optional[ssh.SSHClientFixture]


class AnsiblePlaybookManager(tobiko.SharedFixture):

    def __init__(self):
        super(AnsiblePlaybookManager, self).__init__()
        self._host_playbooks: typing.Dict['AnsiblePlaybookKey',
                                          'AnsiblePlaybook'] = {}

    def get_ansible_playbook(self,
                             ssh_client: ssh.SSHClientType) -> \
            'AnsiblePlaybook':
        ssh_client = ssh.ssh_client_fixture(ssh_client)
        playbook = self._host_playbooks.get(ssh_client)
        if playbook is None:
            playbook = self._init_ansible_playbook(ssh_client=ssh_client)
            self._host_playbooks[ssh_client] = playbook
        return playbook

    def register_ansible_playbook(self, playbook: 'AnsiblePlaybook'):
        ssh_client = ssh.ssh_client_fixture(playbook.ssh_client)
        self._host_playbooks[ssh_client] = playbook

    @staticmethod
    def _init_ansible_playbook(ssh_client: ssh.SSHClientFixture = None) \
            -> 'AnsiblePlaybook':
        if ssh_client is None:
            return local_ansible_playbook()
        else:
            return tobiko.get_fixture(AnsiblePlaybook(ssh_client=ssh_client))


def run_playbook(command: sh.ShellCommand = None,
                 playbook: str = None,
                 playbook_dirname: str = None,
                 playbook_filename: str = None,
                 inventory_filenames: typing.Iterable[str] = None,
                 playbook_files: typing.Iterable[str] = None,
                 roles: typing.Iterable[str] = None,
                 roles_path: typing.Iterable[str] = None,
                 verbosity: int = None,
                 ssh_client: ssh.SSHClientType = None,
                 manager: AnsiblePlaybookManager = None) \
        -> sh.ShellExecuteResult:
    return ansible_playbook(ssh_client=ssh_client,
                            manager=manager).run_playbook(
                            command=command,
                            inventory_filenames=inventory_filenames,
                            playbook=playbook,
                            playbook_dirname=playbook_dirname,
                            playbook_filename=playbook_filename,
                            playbook_files=playbook_files,
                            roles=roles,
                            roles_path=roles_path,
                            verbosity=verbosity)



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack import designate
from tobiko.openstack import keystone
from tobiko.tests.unit import openstack
from tobiko.tests.unit.openstack import test_client


class DesignateClientFixtureTest(test_client.OpenstackClientFixtureTest):

    def create_client(self, session=None):
        return designate.DesignateClientFixture(session=session)


class GetDesignatelientTest(openstack.OpenstackTest):

    def test_get_designate_client(self, session=None, shared=True):
        client1 = designate.get_designate_client(session=session,
                                                 shared=shared)
        client2 = designate.get_designate_client(session=session,
                                                 shared=shared)
        if shared:
            self.assertIs(client1, client2)
        else:
            self.assertIsNot(client1, client2)
        self.assertIsInstance(client1, designate.DESIGNATE_CLIENT_CLASSES)
        self.assertIsInstance(client2, designate.DESIGNATE_CLIENT_CLASSES)

    def test_get_designate_client_with_not_shared(self):
        self.test_get_designate_client(shared=False)

    def test_get_designate_client_with_session(self):
        session = keystone.get_keystone_session()
        self.test_get_designate_client(session=session)


class DesignateClientTest(openstack.OpenstackTest):

    def test_designate_client_with_none(self):
        default_client = designate.get_designate_client()
        client = designate.designate_client(None)
        self.assertIsInstance(client, designate.DESIGNATE_CLIENT_CLASSES)
        self.assertIs(default_client, client)

    def test_designate_client_with_client(self):
        default_client = designate.get_designate_client()
        client = designate.designate_client(default_client)
        self.assertIsInstance(client, designate.DESIGNATE_CLIENT_CLASSES)
        self.assertIs(default_client, client)

    def test_designate_client_with_fixture(self):
        fixture = designate.DesignateClientFixture()
        client = designate.designate_client(fixture)
        self.assertIsInstance(client, designate.DESIGNATE_CLIENT_CLASSES)
        self.assertIs(client, fixture.client)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
TOBIKO_DIR = os.path.abspath(os.path.join(BASE_DIR, "..", ".."))
sys.path.insert(0, TOBIKO_DIR)

# -- Python logging ----------------------------------------------------------

import logging
from tools import common

common.setup_logging(level=logging.INFO)


# -- Project information -----------------------------------------------------

project = 'Tobiko Release Notes'
copyright = "2019, Red Hat"
author = "Tobiko's Team"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#

from tools import get_version
release = get_version.get_version()
version = '.'.join(release.split('.', 2)[:2])


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'reno.sphinxext',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = []

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
try:
    import sphinx_rtd_theme
    html_theme = "sphinx_rtd_theme"
    # Theme options are theme-specific and customize the look and feel of a
    # theme further. For a list of options available for each theme, see the
    # documentation.
    html_theme_options = {
        "logo_only": False,
        "display_version": True,
        "prev_next_buttons_location": "top",
        "style_external_links": True,
        # Toc options
        "collapse_navigation": True,
        "sticky_navigation": True,
        "navigation_depth": 4,
        "includehidden": True,
        "titles_only": False,
    }
except ModuleNotFoundError:
    html_theme = 'alabaster'
    html_theme_options = {}



# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'TobikoReleaseNotesdoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'TobikoReleaseNotes.tex', u'Tobiko Release Notes Documentation',
     u'Tobiko developers', 'manual'),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'tobikoreleasenotes', u'Tobiko Release Notes Documentation',
     [u'Tobiko developers'], 1)
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'TobikoReleaseNotes', u'Tobiko Release Notes Documentation',
     u'Tobiko developers', 'TobikoReleaseNotes', 'One line description of project.',
     'Miscellaneous'),
]



# Copyright (c) 2021 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log
import testtools

import tobiko
from tobiko import config
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import stacks
from tobiko.shell import ip
from tobiko.shell import iperf3
from tobiko.shell import ping
from tobiko.shell import sh
from tobiko.shell import tcpdump


CONF = config.CONF
LOG = log.getLogger(__name__)


@keystone.skip_unless_has_keystone_credentials()
@neutron.skip_if_is_old_ovn()
class QoSNetworkTest(testtools.TestCase):
    """Tests QoS basic functionality"""

    #: Resources stacks with QoS Policy and QoS Rules and Advanced server
    network = tobiko.required_fixture(stacks.QosNetworkStackFixture)
    policy = tobiko.required_fixture(stacks.QosPolicyStackFixture)
    server = tobiko.required_fixture(stacks.QosServerStackFixture)

    def setUp(self):
        super(QoSNetworkTest, self).setUp()
        # this is executed simply to wait until the server boot is completed
        sh.get_hostname(self.server.ssh_client)

    def test_ping_dscp(self):
        capture_file = sh.execute('mktemp', sudo=True).stdout.strip()
        interface = ip.get_network_main_route_device(
            self.server.floating_ip_address)

        # IPv4 tcpdump DSCP filters explanation:
        # ip[1] refers to the byte 1 (the TOS byte) of the IP header
        # 0xfc = 11111100 is the mask to get only DSCP value from the ToS
        # As DSCP mark is most significant 6 bits we do right shift (>>)
        # twice in order to divide by 4 and compare with the decimal value
        # See details at http://darenmatthews.com/blog/?p=1199
        dscp_mark = CONF.tobiko.neutron.dscp_mark
        capture_filter = (f"'(ip src {self.server.floating_ip_address} and "
                          f"(ip[1] & 0xfc) >> 2 == {dscp_mark})'")

        # start a capture
        process = tcpdump.start_capture(
            capture_file=capture_file,
            interface=interface,
            capture_filter=capture_filter,
            capture_timeout=60)
        # send a ping to the server
        ping.assert_reachable_hosts([self.server.floating_ip_address], count=5)
        # stop tcpdump and get the pcap capture
        pcap = tcpdump.get_pcap(process, capture_file=capture_file)
        # check the capture is not empty
        tcpdump.assert_pcap_is_not_empty(pcap=pcap)

    def test_network_qos_policy_id(self):
        """Verify network policy ID"""
        self.assertEqual(self.policy.qos_policy_id,
                         self.network.qos_policy_id)

    def test_server_qos_policy_id(self):
        """Verify server policy ID"""
        self.assertIsNone(self.server.port_details['qos_policy_id'])

    def test_qos_bw_limit(self):
        """Verify BW limit using the iperf3 tool"""
        self.server.wait_for_iperf3_server()
        # localhost will act as client
        bandwidth_limit = self.policy.bwlimit_kbps * 1000.
        for attempt in tobiko.retry(timeout=100., interval=5.):
            try:
                iperf3.assert_has_bandwith_limits(
                    address=self.server.ip_address,
                    min_bandwith=bandwidth_limit * 0.85,
                    max_bandwith=bandwidth_limit * 1.15,
                    port=self.server.iperf3_port,
                    download=True)
                break
            except sh.ShellCommandFailed as err:
                if ('unable to connect to server: Connection refused'
                        in str(err)):
                    attempt.check_limits()
                    LOG.debug('iperf command failed because the iperf server '
                              'was not ready yet - retrying...')
                else:
                    raise err



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import tobiko
from tobiko.openstack.neutron import _client
from tobiko.openstack.neutron import _network
from tobiko.openstack.neutron import _port


FloatingIpType = typing.Dict[str, typing.Any]
FloatingIpIdType = typing.Union[str, FloatingIpType]


def get_floating_ip_id(floating_ip: FloatingIpIdType) -> str:
    if isinstance(floating_ip, str):
        return floating_ip
    else:
        return floating_ip['id']


def get_floating_ip(floating_ip: FloatingIpType,
                    client: _client.NeutronClientType = None,
                    **params):
    floating_ip_id = get_floating_ip_id(floating_ip)
    try:
        return _client.neutron_client(client).show_floatingip(
            floating_ip_id, **params)['floatingip']
    except _client.NotFound as ex:
        raise NoSuchFloatingIp(id=floating_ip_id) from ex


def find_floating_ip(client: _client.NeutronClientType = None,
                     unique=False,
                     default: FloatingIpType = None,
                     **params) -> FloatingIpType:
    """Look for a port matching some property values"""
    floating_ips = list_floating_ips(client=client, **params)
    if default is None or floating_ips:
        if unique:
            return floating_ips.unique
        else:
            return floating_ips.first
    else:
        return default


def list_floating_ips(client: _client.NeutronClientType = None,
                      retrieve_all=True,
                      port: _port.PortIdType = None,
                      **params) -> tobiko.Selection[FloatingIpType]:
    if port is not None:
        params['port_id'] = _port.get_port_id(port)
    floating_ips = _client.neutron_client(client).list_floatingips(
        retrieve_all=retrieve_all, **params)['floatingips']
    return tobiko.select(floating_ips)


def create_floating_ip(network: _network.NetworkIdType = None,
                       port: _port.PortIdType = None,
                       client: _client.NeutronClientType = None,
                       add_cleanup=True,
                       **params) -> FloatingIpType:
    if network is None:
        network = params.get('floating_network_id')
        if network is None:
            from tobiko.openstack import stacks
            network = stacks.get_floating_network_id()
    params['floating_network_id'] = _network.get_network_id(network)
    if port is not None:
        params['port_id'] = _port.get_port_id(port)
    floating_ip: FloatingIpType = _client.neutron_client(
        client).create_floatingip(body={'floatingip': params})['floatingip']
    if add_cleanup:
        tobiko.add_cleanup(cleanup_floating_ip,
                           floating_ip=floating_ip,
                           client=client)
    return floating_ip


def cleanup_floating_ip(floating_ip: FloatingIpIdType,
                        client: _client.NeutronClientType = None):
    try:
        delete_floating_ip(floating_ip=floating_ip,
                           client=client)
    except NoSuchFloatingIp:
        pass


def delete_floating_ip(floating_ip: FloatingIpIdType,
                       client: _client.NeutronClientType = None):
    floating_ip_id = get_floating_ip_id(floating_ip)
    try:
        _client.neutron_client(client).delete_floatingip(floating_ip_id)
    except _client.NotFound as ex:
        raise NoSuchFloatingIp(id=floating_ip_id) from ex


def update_floating_ip(floating_ip: FloatingIpIdType,
                       client: _client.NeutronClientType = None,
                       **params) -> FloatingIpType:
    floating_ip_id = get_floating_ip_id(floating_ip)
    try:
        return _client.neutron_client(client).update_floatingip(
                floating_ip_id, body={'floatingip': params})['floatingip']
    except _client.NotFound as ex:
        raise NoSuchFloatingIp(id=floating_ip_id) from ex


class NoSuchFloatingIp(tobiko.ObjectNotFound):
    message = "No such floating IP found for {id!r}"



# Copyright 2023 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing
import time

from oslo_log import log
from manilaclient import exceptions

import tobiko
from tobiko.openstack.manila import _client
from tobiko.openstack.manila import _constants
from tobiko.openstack.manila import _exceptions

LOG = log.getLogger(__name__)


def wait_for_status(object_id: str,
                    status_key: str = _constants.RESOURCE_STATUS,
                    status: str = _constants.STATUS_AVAILABLE,
                    get_client: typing.Callable = None,
                    interval: tobiko.Seconds = None,
                    timeout: tobiko.Seconds = None,
                    **kwargs):
    """Waits for an object to reach a specific status.

    :param status_key: The key of the status field in the response.
                       Ex. status
    :param status: The status to wait for. Ex. "ACTIVE"
    :param get_client: The tobiko client get method.
                        Ex. _client.get_zone
    :param object_id: The id of the object to query.
    :param interval: How often to check the status, in seconds.
    :param timeout: The maximum time, in seconds, to check the status.
    :raises TimeoutException: The object did not achieve the status or ERROR in
                              the check_timeout period.
    :raises UnexpectedStatusException: The request returned an unexpected
                                       response code.
    """

    get_client = get_client or _client.get_share

    for attempt in tobiko.retry(timeout=timeout,
                                interval=interval,
                                default_timeout=300.,
                                default_interval=5.):
        response = get_client(object_id, **kwargs)
        if response[status_key] == status:
            return response

        attempt.check_limits()

        LOG.debug(f"Waiting for {get_client.__name__} {status_key} to get "
                  f"from '{response[status_key]}' to '{status}'...")


def wait_for_share_status(share_id):
    wait_for_status(object_id=share_id)


def _is_share_deleted(share_id):
    try:
        res = _client.get_share(share_id)
    except _exceptions.ShareNotFound:
        return True
    if res.get(_constants.RESOURCE_STATUS) in [
            _constants.STATUS_ERROR, _constants.STATUS_ERROR_DELETING]:
        # Share has "error_deleting" status and can not be deleted.
        raise _exceptions.ShareReleaseFailed(id=share_id)
    return False


def wait_for_resource_deletion(share_id, build_interval=1, build_timeout=60):
    """Waits for a resource to be deleted."""
    start_time = int(time.time())
    while True:
        if _is_share_deleted(share_id):
            return
        if int(time.time()) - start_time >= build_timeout:
            raise exceptions.TimeoutException
        time.sleep(build_interval)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
from urllib import parse
import typing

from oslo_log import log

from tobiko.docker import _exception
from tobiko.shell import sh
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


def discover_docker_urls(
        ssh_client: ssh.SSHClientType = None,
        default_url='unix:/var/run/docker.sock',
        check_urls=True,
        sudo=False) \
        -> typing.List[str]:
    last_error: typing.Optional[Exception] = None
    urls = []
    try:
        urls += list_docker_urls(ssh_client=ssh_client,
                                 sudo=sudo)
    except _exception.DockerUrlNotFoundError as ex:
        last_error = ex
    if default_url not in urls:
        urls.append(default_url)
    if check_urls:
        urls = select_valid_urls(ssh_client=ssh_client,
                                 urls=urls,
                                 sudo=sudo,
                                 last_error=last_error)
    return urls


def list_docker_urls(
        ssh_client: ssh.SSHClientType,
        sudo=False) \
        -> typing.List[str]:
    urls = []
    try:
        processes = sh.list_processes(command='^dockerd',
                                      ssh_client=ssh_client,
                                      sudo=sudo)
    except sh.PsError as ex:
        raise _exception.DockerUrlNotFoundError(
            "Docker demon is not running") from ex

    for process in processes:
        if process.command_line:
            urls += urls_from_command_line(process.command_line)
    urls = list(collections.OrderedDict.fromkeys(urls))
    return urls


def select_valid_urls(
        ssh_client: ssh.SSHClientType,
        urls: typing.List[str],
        sudo=False,
        last_error: Exception = None) \
        -> typing.List[str]:
    valid_urls = []
    for url in urls:
        parsed_url = parse.urlparse(url)
        if (parsed_url.scheme == 'unix' and
                parsed_url.path.startswith('/')):
            try:
                sh.execute(f"test -r '{parsed_url.path}'",
                           ssh_client=ssh_client,
                           sudo=sudo)
            except sh.ShellCommandFailed as ex:
                LOG.exception(
                    f"Can't read from socket: {parsed_url.path}")
                ex.__cause__ = last_error
                last_error = ex
            else:
                valid_urls.append(url)

    if not valid_urls:
        raise _exception.DockerUrlNotFoundError(
            "Docker is not running") from last_error
    return valid_urls


def urls_from_command_line(command_line: sh.ShellCommand,
                           default_url='unix:/var/run/docker.sock') \
        -> typing.List[str]:
    urls = []
    arg_is_url = False
    for arg in command_line[1:]:
        if arg == '-H':
            arg_is_url = True
            continue
        if arg_is_url:
            arg_is_url = False
            try:
                url = parse.urlparse(arg)
            except Exception:
                LOG.debug(f'Invalid URL: {arg}', exc_info=1)
                continue
            if url.scheme == 'fd':
                urls.append(default_url)
                continue
            if url.scheme != 'unix':
                LOG.debug(f'Unsupported URL scheme: {arg}')
                continue
            if not url.path.startswith('/'):
                LOG.debug(f'Unsupported URL path: {arg}')
                continue
            urls.append(arg)
    return urls


def is_docker_running(ssh_client=None, sudo=False, **execute_params):
    try:
        discover_docker_urls(ssh_client=ssh_client,
                             sudo=sudo,
                             **execute_params)
    except _exception.DockerUrlNotFoundError:
        return False
    else:
        return True



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from collections import abc
import typing

from tobiko.openstack.designate import _client

DesignateZone = typing.Mapping[str, typing.Any]
DesignateZoneType = typing.Union[str, typing.Mapping[str, typing.Any]]


def designate_zone_id(zone: DesignateZoneType) -> str:
    if isinstance(zone, str):
        return zone
    elif isinstance(zone, abc.Mapping):
        return zone['id']
    else:
        raise TypeError(f'{zone} object is an invalid Designate zone type')


def get_designate_zone(zone: str,
                       client: _client.DesignateClientType = None) \
        -> DesignateZone:
    zone_id = designate_zone_id(zone)
    return _client.designate_client(client).zones.get(zone_id)



# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import argparse
import json
import logging
import os
import subprocess
import sys


LOG = logging.getLogger(__name__)

GET_PYTHON_VERSION_SCRIPT = """
import sys

version = '.'.join(str(i) for i in sys.version_info)
print(version)
"""


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--quiet', '-q',
                        action="store_true",
                        help="mute logging messages from STDERR")
    parser.add_argument('--base', '-b',
                        action="store_true",
                        help="print Python base prefix")
    args = parser.parse_args()

    setup_logging(quiet=args.quiet)
    version = get_python_version()
    executables = [
        info['executable']
        for info in iter_python_executables_info(match_version=version,
                                                 base=args.base)]
    info = {'version': version,
            'executable': sys.executable,
            'executables': executables}
    output = json.dumps(info)
    assert output.startswith('{')
    print(output)


def setup_logging(quiet=False):
    if quiet:
        level = logging.ERROR
    else:
        level = logging.DEBUG
    logging.basicConfig(
        level=level,
        stream=sys.stderr,
        format='%(name)-s: %(levelname)-7s %(asctime)-15s | %(message)s')


def get_python_version():
    return '.'.join(str(i) for i in sys.version_info)


def iter_python_executables_info(match_version=None, base=None):
    last_error = None
    for executable in iter_python_executables(base=base):
        command = subprocess.list2cmdline(
            [executable, '-c', GET_PYTHON_VERSION_SCRIPT])
        try:
            version = execute(command).splitlines()[0]
        except subprocess.CalledProcessError:
            LOG.exception('Unable to get version from script')
        else:
            if not match_version or match_version == version:
                yield {'executable': executable, 'version': version}
    else:
        if last_error:
            raise last_error


def iter_python_executables(base):
    if base:
        prefix = getattr(sys, 'base_prefix', sys.prefix)
    else:
        prefix = sys.prefix
    if prefix:
        if os.path.isdir(prefix):
            for python_name in iter_versioned_names():
                executable = os.path.join(prefix, 'bin', python_name)
                if os.path.isfile(executable):
                    yield executable


def iter_versioned_names(unversioned=None):
    unversioned = unversioned or 'python'
    short_versioned = unversioned + str(sys.version_info[0])
    long_versioned = '.'.join([short_versioned, str(sys.version_info[1])])
    yield long_versioned
    yield short_versioned
    yield unversioned


def execute(command, *args, **kwargs):
    if args or kwargs:
        command = command.format(*args, **kwargs)
    LOG.debug('%s', command)
    env = kwargs.get('env', None)
    return subprocess.check_output(command, shell=True,
                                   universal_newlines=True,
                                   env=env)


def name_from_path(path):
    return os.path.splitext(os.path.basename(path))[0]


if __name__ == '__main__':
    LOG = logging.getLogger(name_from_path(__file__))
    main()



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing


class TcpdumpParameters(typing.NamedTuple):
    capture_file: str
    interface: typing.Optional[str] = None
    capture_filter: typing.Optional[str] = None
    capture_timeout: typing.Optional[int] = None


def tcpdump_parameters(
        capture_file: str,
        interface: str = None,
        capture_filter: str = None,
        capture_timeout: int = None):
    """Get tcpdump parameters
    """
    return TcpdumpParameters(capture_file=capture_file,
                             interface=interface,
                             capture_filter=capture_filter,
                             capture_timeout=capture_timeout)



# Copyright (c) 2022 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import contextlib
import io
import multiprocessing
import os
import signal
import typing

from oslo_log import log

from tobiko import _exception
from tobiko import _fixture
from tobiko import _retry
from tobiko import _time


LOG = log.getLogger(__name__)


@contextlib.contextmanager
def pause_background_process(target: typing.Callable,
                             *args,
                             process_name: str = None,
                             pid_file: str = None,
                             retry_timeout: _time.Seconds = None,
                             retry_interval: _time.Seconds = None,
                             **kwargs):
    """It stops background process (if running), finally it restart it
    """
    if process_name is None:
        process_name = _fixture.get_object_name(target)
    if pid_file is None:
        pid_file = get_pid_file(process_name)
    stop_background_process(pid_file=pid_file,
                            retry_timeout=retry_timeout,
                            retry_interval=retry_interval)
    try:
        yield
    finally:
        start_background_process(target,
                                 *args,
                                 process_name=process_name,
                                 pid_file=pid_file,
                                 retry_timeout=retry_timeout,
                                 retry_interval=retry_interval,
                                 **kwargs)


class StartBackgroundProcessError(_exception.TobikoException):
    message = ("Error starting background process:\n"
               "  target: {target_name}\n"
               "  process name: {process_name}\n"
               "  PID file: {pid_file}\n"
               "  args: {args}\n"
               "  kwargs: {kwargs}\n"
               "  reason: {reason}\n")


def start_background_process(target: typing.Callable,
                             *args,
                             process_name: str = None,
                             pid_file: str = None,
                             retry_timeout: _time.Seconds = None,
                             retry_interval: _time.Seconds = None,
                             max_pids: typing.Optional[int] = 1,
                             **kwargs) -> bool:
    target_name = _fixture.get_object_name(target)
    if process_name is None:
        process_name = target_name
    if pid_file is None:
        pid_file = get_pid_file(process_name)

    initial_pids = check_background_process(pid_file=pid_file)
    if max_pids is not None:
        if len(initial_pids) >= max_pids:
            LOG.debug('Process already started:\n'
                      f"  target: {target_name}\n"
                      f"  process name: {process_name}\n"
                      f"  PID file: {pid_file}\n"
                      f"  PIDs: {initial_pids}\n"
                      f"  max PIDs: {max_pids}\n")
            return False

    LOG.debug('Starting parent background process...\n'
              f"  target: {target_name}\n"
              f"  process name: {process_name}\n"
              f"  PID file: {pid_file}\n"
              f"  args: {args}\n"
              f"  kwargs {kwargs}\n")
    # start parent process, nested with a started child process
    # then kill the parent
    parameters = BackgroundProcessParameters(target=target,
                                             args=args,
                                             kwargs=kwargs,
                                             process_name=process_name,
                                             pid_file=pid_file,
                                             max_pids=max_pids)
    process = multiprocessing.Process(target=_run_parent_background_process,
                                      name=process_name,
                                      args=(parameters,),
                                      daemon=False)
    process.start()
    LOG.debug(f'Parent background process started (PID={process.pid}).')
    try:
        for attempt in _retry.retry(timeout=retry_timeout,
                                    interval=retry_interval,
                                    default_timeout=15.,
                                    default_interval=.5):
            final_pids = check_background_process(pid_file=pid_file)
            new_pids = sorted(set(final_pids) - set(initial_pids))
            if new_pids:
                LOG.debug(f'New background process started (PIDs={new_pids})')
                break
            if attempt.is_last:
                raise StartBackgroundProcessError(target_name=target_name,
                                                  process_name=process_name,
                                                  pid_file=pid_file,
                                                  args=args,
                                                  kwargs=kwargs,
                                                  reason="timed out")
    finally:
        process.terminate()
        LOG.debug('Background process orphaned')
    return True


def _run_parent_background_process(parameters: 'BackgroundProcessParameters'):
    target_name = _fixture.get_object_name(parameters.target)
    LOG.debug('Starting background process...\n'
              f'  target: {target_name}\n'
              f"  process name: {parameters.process_name}\n"
              f"  PID file: {parameters.pid_file}\n"
              f"  args: {parameters.args}\n"
              f"  kwargs {parameters.kwargs}\n")
    process = multiprocessing.Process(target=_run_background_process,
                                      name=parameters.process_name,
                                      args=(parameters,))
    process.start()
    LOG.debug('Background process started\n'
              f'  target: {target_name}\n'
              f"  process name: {parameters.process_name}\n"
              f"  PID file: {parameters.pid_file}\n"
              f"  PID: {process.pid}\n"
              f"  args: {parameters.args}\n"
              f"  kwargs {parameters.kwargs}\n")
    process.join()


def _run_background_process(parameters: 'BackgroundProcessParameters'):
    # In order to ensure max_pids counter is respected I must check
    target_name = _fixture.get_object_name(parameters.target)
    if parameters.max_pids is not None:
        pids = check_background_process(pid_file=parameters.pid_file)
        if len(pids) >= parameters.max_pids:
            LOG.debug('Aborting background process execution:\n'
                      f"  target: {target_name}\n"
                      f"  process name: {parameters.process_name}\n"
                      f"  PID file: {parameters.pid_file}\n"
                      f"  PIDs: {pids}\n"
                      f"  max PIDs: {parameters.max_pids}\n")
            raise RuntimeError("Background job execution aborted")

    pid = os.getpid()
    with open_pid_file(parameters.pid_file, "at") as fd:
        fd.write(f"{pid}\n")

    assert not is_background_process()
    set_background_process_parameters(parameters)
    assert is_background_process()
    LOG.info(f'Background process is running:\n:'
             f'  target: {target_name}\n'
             f'  name: {parameters.process_name}\n'
             f'  PID file: {parameters.pid_file}\n'
             f'  PID: {pid}\n'
             f"  args: {parameters.args}\n"
             f"  kwargs {parameters.kwargs}\n")
    try:
        result = parameters.target(*parameters.args, **parameters.kwargs)
    except Exception:
        LOG.exception(f'Background process is failed:\n:'
                      f'  target: {target_name}\n'
                      f'  name: {parameters.process_name}\n'
                      f'  PID file: {parameters.pid_file}\n'
                      f'  PID: {pid}\n'
                      f"  args: {parameters.args}\n"
                      f"  kwargs: {parameters.kwargs}\n")
        raise
    else:
        LOG.info(f'Background process has terminated:\n:'
                 f'  target: {target_name}\n'
                 f'  name: {parameters.process_name}\n'
                 f'  PID file: {parameters.pid_file}\n'
                 f'  PID: {pid}\n'
                 f"  args: {parameters.args}\n"
                 f"  kwargs: {parameters.kwargs}\n"
                 f"  result: {result}\n")
        return result


def stop_background_process(*pids: int,
                            pid_file: str = None,
                            terminate_signal=signal.SIGTERM,
                            kill_signal=signal.SIGKILL,
                            retry_timeout: _time.Seconds = None,
                            retry_interval: _time.Seconds = None):
    signaled_pids = signal_background_process(*pids,
                                              pid_file=pid_file,
                                              signal_number=terminate_signal)
    if pid_file is not None:
        remove_pid_file(pid_file)
    if signaled_pids:
        try:
            for attempt in _retry.retry(timeout=retry_timeout,
                                        interval=retry_interval,
                                        default_timeout=15.,
                                        default_interval=.5):
                signaled_pids = check_background_process(*signaled_pids)
                if attempt.is_last or not signaled_pids:
                    break
                LOG.debug(f"Waiting for processes to terminate: "
                          f"{signaled_pids}...")
        finally:
            signal_background_process(*signaled_pids,
                                      signal_number=kill_signal)


def read_pid_file(pid_file: str) -> typing.Tuple[int, ...]:
    """list PIDs of specified bg_process_name file"""
    pids = []
    try:
        fd = open_pid_file(pid_file=pid_file, mode='rt')
    except FileNotFoundError:
        pass
    else:
        with fd:
            for line_number, line in enumerate(fd.readlines(), start=1):
                line = line.strip()
                pid = None
                if line:
                    try:
                        pid = int(line.rstrip())
                    except (TypeError, ValueError):
                        LOG.exception(f"{pid_file}:{line_number}: value is "
                                      f"not an integer ({line}).")
                        continue
                if not pid:
                    continue
                pids.append(pid)
    return tuple(pids)


def get_pid_file(process_name: str) -> str:
    return os.path.expanduser(f'~/.tobiko/pid/{process_name}.pid')


def remove_pid_file(pid_file: str):
    try:
        os.remove(pid_file)
    except FileNotFoundError:
        pass
    else:
        LOG.debug(f"PID file '{pid_file}' removed")


def open_pid_file(pid_file: str, mode: str):
    if set('aw') & set(mode):
        # Ensure pids directory exists
        os.makedirs(os.path.dirname(pid_file), exist_ok=True)
    return io.open(pid_file, mode)


def check_background_process(*pids: int,
                             pid_file: str = None) -> typing.Tuple[int, ...]:
    return signal_background_process(*pids, pid_file=pid_file, signal_number=0)


def signal_background_process(*pids: int,
                              pid_file: str = None,
                              signal_number: int = signal.SIGTERM) \
        -> typing.Tuple[int, ...]:
    if pid_file:
        pids += read_pid_file(pid_file)
    if not pids:
        return tuple()

    signaled_pids: typing.List[int] = []
    for pid in pids:
        try:
            os.kill(pid, signal_number)
            signaled_pids.append(pid)
        except ProcessLookupError:
            pass

    LOG.debug(f"Signal {signal_number} sent to process (PIDs={pids})")
    return tuple(signaled_pids)


def is_background_process() -> bool:
    return get_background_process_parameters() is not None


class BackgroundProcessParameters(typing.NamedTuple):
    target: typing.Callable
    args: typing.Tuple
    kwargs: typing.Dict[str, typing.Any]
    process_name: str
    pid_file: str
    max_pids: typing.Optional[int]


BACKGROUND_PROCESS_PARAMETERS: \
    typing.Optional[BackgroundProcessParameters] = None


def get_background_process_parameters() \
        -> typing.Optional[BackgroundProcessParameters]:
    return BACKGROUND_PROCESS_PARAMETERS


def set_background_process_parameters(parameters: BackgroundProcessParameters):
    # pylint: disable=global-statement
    global BACKGROUND_PROCESS_PARAMETERS
    BACKGROUND_PROCESS_PARAMETERS = parameters


class BackgroundProcessFixture(_fixture.SharedFixture):

    def __init__(self,
                 *args,
                 target: typing.Callable = None,
                 process_name: str = None,
                 pid_file: str = None,
                 retry_timeout: _time.Seconds = None,
                 retry_interval: _time.Seconds = None,
                 terminate_signal=signal.SIGTERM,
                 kill_signal=signal.SIGTERM,
                 **kwargs):
        super().__init__()
        if target is None:
            target = self.run
            if process_name is None:
                process_name = _fixture.get_fixture_name(self)
        elif process_name is None:
            process_name = _fixture.get_object_name(target)
        if pid_file is None:
            pid_file = get_pid_file(process_name)
        self._target = target
        self._process_name = process_name
        self._pid_file = pid_file
        self._target = target
        self._args = args
        self._kwargs = kwargs
        self._retry_timeout = retry_timeout
        self._retry_interval = retry_interval
        self._terminate_signal = terminate_signal
        self._kill_signal = kill_signal

    def setup_fixture(self):
        if not is_background_process():
            start_background_process(type(self)._run,
                                     *self._args,
                                     process_name=self._process_name,
                                     pid_file=self._pid_file,
                                     retry_timeout=self._retry_timeout,
                                     retry_interval=self._retry_interval,
                                     **self._kwargs)

    def cleanup_fixture(self):
        stop_background_process(pid_file=self._pid_file,
                                terminate_signal=self._terminate_signal,
                                kill_signal=self._kill_signal,
                                retry_timeout=self._retry_timeout,
                                retry_interval=self._retry_interval)

    @contextlib.contextmanager
    def pause(self):
        return pause_background_process(target=self._target,
                                        process_name=self._process_name,
                                        pid_file=self._pid_file,
                                        args=self._args,
                                        kwargs=self._kwargs,
                                        retry_timeout=self._retry_timeout,
                                        retry_interval=self._retry_interval)

    def start(self):
        _fixture.setup_fixture(self)

    def stop(self):
        _fixture.cleanup_fixture(self)

    def kill(self):
        signal_background_process(pid_file=self._pid_file,
                                  signal_number=self._kill_signal)
        _fixture.cleanup_fixture(self)

    def check(self):
        return check_background_process(pid_file=self._pid_file)

    @property
    def is_alive(self) -> bool:
        return bool(self.check())

    @classmethod
    def _run(cls, *args, **kwargs):
        # pylint: disable=protected-access
        assert is_background_process()
        return _fixture.setup_fixture(cls)._target(*args, **kwargs)

    def run(self, *args, **kwargs):
        raise NotImplementedError



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import json
import re
import typing

import netaddr
import pytest
from neutron_lib import constants
from oslo_log import log
import testtools

import tobiko
from tobiko.shell import ping
from tobiko.shell import ip
from tobiko.shell import sh
from tobiko.openstack import neutron
from tobiko.openstack.neutron import neutron_log_reader
from tobiko.openstack import stacks
from tobiko.openstack import topology
from tobiko.tripleo import overcloud


LOG = log.getLogger(__name__)
IPV4 = constants.IP_VERSION_4
IPV6 = constants.IP_VERSION_6


@pytest.mark.minimal
class PortTest(testtools.TestCase):
    """Test Neutron ports"""

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_fixture(stacks.CirrosServerStackFixture)

    def test_port_ips(self, ip_version: typing.Optional[int] = None):
        """Checks port IPS has been assigned to server via DHCP protocol"""
        port_ips = set(neutron.list_device_ip_addresses(
            device=self.stack.server_id,
            network=self.stack.network_stack.network_id,
            need_dhcp=self.stack.need_dhcp,
            ip_version=ip_version))
        if port_ips:
            # verify neutron port IPs and VM port IPs match
            # when a VM connected to the external network has been just
            # created, it may need some time to receive its IPv6 address
            for attempt in tobiko.retry(timeout=60., interval=4.):
                server_ips = set(ip.list_ip_addresses(
                    scope='global', ssh_client=self.stack.ssh_client))
                server_ips &= port_ips  # ignore other server IPs
                LOG.debug("Neutron IPs and VM IPs should match...")
                try:
                    self.assertEqual(
                        port_ips, server_ips,
                        f"Server {self.stack.server_id} is missing port "
                        f"IP(s): {port_ips - server_ips}")
                    break
                except self.failureException:
                    attempt.check_limits()
        elif ip_version:
            self.skipTest(f"Server has any port IPv{ip_version} address to be"
                          " tested")
        else:
            self.skipTest("Server has any port IP address to be tested")

    def test_port_network(self):
        self.assertEqual(self.stack.network_stack.network_id,
                         self.stack.port_details['network_id'])

    def test_port_subnets(self):
        """Checks port subnets"""
        port_subnets = [fixed_ip['subnet_id']
                        for fixed_ip in self.stack.port_details['fixed_ips']]
        network_subnets = self.stack.network_stack.network_details['subnets']
        self.assertEqual(set(network_subnets), set(port_subnets))

    def test_ping_subnet_gateways(self):
        """Checks server can ping its gateway IPs"""
        network_id = self.stack.network_stack.network_id
        subnets = neutron.list_subnets(network_id=network_id)
        LOG.debug("Subnets enabled are:\n"
                  f"{json.dumps(subnets, indent=4, sort_keys=True)}")
        gateway_ips = [netaddr.IPAddress(subnet['gateway_ip'])
                       for subnet in subnets]
        LOG.debug(f"Gateway IPs are: {gateway_ips}")
        ping.assert_reachable_hosts(gateway_ips,
                                    ssh_client=self.stack.ssh_client)

    def test_ping_port(self, network_id=None, device_id=None, ip_version=None):
        """Checks server can ping its own port"""
        port_ips = neutron.list_device_ip_addresses(
            device=device_id or self.stack.server_id,
            network=network_id or self.stack.network_stack.network_id,
            need_dhcp=self.stack.need_dhcp,
            ip_version=ip_version)
        if port_ips:
            ping.assert_reachable_hosts(port_ips,
                                        timeout=600.,
                                        ssh_client=self.stack.peer_ssh_client)
        elif ip_version:
            self.skipTest(f"Server has any port IPv{ip_version} address to be"
                          " tested")
        else:
            self.skipTest("Server has any port IP address to be tested")


# --- Test opening ports on external network ----------------------------------

@stacks.skip_unless_has_external_network
class UbuntuExternalPortTest(PortTest):
    """Test Neutron ports"""

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_fixture(
        stacks.UbuntuExternalServerStackFixture)


# --- Test la-h3 extension ----------------------------------------------------

@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(binary='neutron-l3-agent',
                                           count=2)
class L3HAPortTest(PortTest):
    #: Resources stack with floating IP and Nova server
    stack = tobiko.required_fixture(stacks.L3haServerStackFixture)


class ExtraDhcpOptsPortTest(PortTest):
    """Test extra-dhcp-options port parameter"""
    stack = tobiko.required_fixture(
        stacks.ExtraDhcpOptsCirrosServerStackFixture)

    def test_extra_dhcp_opts(self):
        port = neutron.get_port(self.stack.port_id)
        for option in port['extra_dhcp_opts']:
            if 'domain-name' == option['opt_name']:
                domain = option['opt_value'].replace('"', '')
                break
        else:
            tobiko.fail('No extra-dhcp-opt found for domain-name')

        vm_resolv_conf = sh.execute('cat /etc/resolv.conf',
                                    ssh_client=self.stack.ssh_client).stdout
        self.assertIsNotNone(
            re.search(r'^search\s+{domain}$'.format(domain=domain),
                      vm_resolv_conf,
                      re.MULTILINE))


@overcloud.skip_unless_ovn_bgp_agent
class NoFipPortTest(ExtraDhcpOptsPortTest):
    stack = tobiko.required_fixture(stacks.CirrosNoFipServerStackFixture)

    def setUp(self):
        super().setUp()

        try:
            node = topology.find_openstack_node(group='networker')
        except topology.NoSuchOpenStackTopologyNodeGroup:
            node = topology.find_openstack_node(group='controller')
        expose_tenant_networks = topology.get_config_setting(
            'bgp-agent.conf', node.ssh_client, 'expose_tenant_networks')
        if expose_tenant_networks is None or (
                expose_tenant_networks.lower() != 'true'):
            tobiko.skip_test('BGP expose_tenant_networks is disabled')


@neutron.skip_unless_is_ovn()
class ExtraDhcpOptsPortLoggingTest(testtools.TestCase):

    stack = tobiko.required_fixture(stacks.NetworkStackFixture)

    def test_extra_dhcp_opts_logs_unsupported_options(self):
        # initialize logs that match the pattern
        neutron_log_reader.assert_ovn_unsupported_dhcp_option_messages()

        wrong_ipv4_option = 'wrong-ipv4-option'
        wrong_ipv6_option = 'bananas'
        a_valid_ipv4_option_used_for_ipv6 = 'log-server'
        extra_dhcp_opts = [
            {'opt_value': '1.1.1.1',
             'opt_name': a_valid_ipv4_option_used_for_ipv6,
             'ip_version': IPV6},
            {'opt_value': 'ipv6.domain',
             'opt_name': 'domain-search',
             'ip_version': IPV6},
            {'opt_value': '1600',
             'opt_name': 'mtu',
             'ip_version': IPV4},
            {'opt_value': 'blablabla',
             'opt_name': wrong_ipv4_option,
             'ip_version': IPV4}]
        # create port with extra-dhcp-opts
        port = neutron.create_port(**{'network_id': self.stack.network_id,
                                      'extra_dhcp_opts': extra_dhcp_opts})
        self.addCleanup(neutron.delete_port, port['id'])
        # find new logs that match the pattern
        invalid_options = [wrong_ipv4_option,
                           a_valid_ipv4_option_used_for_ipv6]
        # assert every invalid dhcp option is logged
        neutron_log_reader.assert_ovn_unsupported_dhcp_option_messages(
            unsupported_options=invalid_options,
            port_uuid=port['id'])

        extra_dhcp_opts.append({'opt_value': '1.1.1.1',
                                'opt_name': wrong_ipv6_option,
                                'ip_version': IPV6})
        # update port with new extra-dhcp-opts
        port = neutron.update_port(port['id'],
                                   **{'extra_dhcp_opts': extra_dhcp_opts})
        invalid_options.append(wrong_ipv6_option)
        # assert every invalid dhcp option is logged
        neutron_log_reader.assert_ovn_unsupported_dhcp_option_messages(
            unsupported_options=invalid_options,
            port_uuid=port['id'])



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko
from tobiko import config
from tobiko.openstack import heat
from tobiko.openstack import neutron
from tobiko.openstack.stacks import _neutron
from tobiko.openstack.stacks import _hot
from tobiko.openstack.stacks import _ubuntu


CONF = config.CONF


@neutron.skip_if_missing_networking_extensions('qos')
class QosPolicyStackFixture(heat.HeatStackFixture):
    """Heat stack with a QoS Policy and some QoS Policy Rules
    """
    has_qos_policy = True
    has_bwlimit = True
    has_dscp_marking = True
    bwlimit_kbps = CONF.tobiko.neutron.bwlimit_kbps
    bwlimit_burst_kbps = int(0.8 * bwlimit_kbps)
    direction = CONF.tobiko.neutron.direction
    dscp_mark = CONF.tobiko.neutron.dscp_mark

    #: Heat template file
    template = _hot.heat_template_file('neutron/qos.yaml')


@neutron.skip_if_missing_networking_extensions('qos')
class QosNetworkStackFixture(_neutron.NetworkBaseStackFixture):

    #: stack with the qos policy for the network
    qos_stack = tobiko.required_fixture(QosPolicyStackFixture)

    has_qos_policy = True

    @property
    def network_value_specs(self):
        value_specs = super().network_value_specs
        return dict(value_specs, qos_policy_id=self.qos_stack.qos_policy_id)

    @tobiko.interworker_synched('create_qos_network_stack')
    def setup_stack(self):
        super().setup_stack()


class QosServerStackFixture(_ubuntu.UbuntuServerStackFixture):
    #: stack with the network with a qos policy
    network_stack = tobiko.required_fixture(QosNetworkStackFixture)

    @property
    def has_vlan(self) -> bool:
        # Trunk ports are not supported with QoS when ml2/ovs is used
        return False



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import


def setup_tobiko_config(conf):
    # pylint: disable=unused-argument
    from tobiko.tripleo import _ansible
    from tobiko.tripleo import overcloud
    from tobiko.tripleo import topology

    _ansible.setup_undercloud_ansible_playbook()
    overcloud.setup_overcloud_keystone_credentials()
    topology.setup_tripleo_topology()



# Copyright (c) 2021 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Octavia attributes
PROVISIONING_STATUS = 'provisioning_status'
OPERATING_STATUS = 'operating_status'

# Octavia provisioning and operating status
ACTIVE = 'ACTIVE'
ERROR = 'ERROR'
PENDING_UPDATE = 'PENDING_UPDATE'
ONLINE = 'ONLINE'

# Octavia protocols
PROTOCOL_HTTP = 'HTTP'
PROTOCOL_TCP = 'TCP'

# Octavia lb algorithms
LB_ALGORITHM_ROUND_ROBIN = 'ROUND_ROBIN'
LB_ALGORITHM_SOURCE_IP_PORT = 'SOURCE_IP_PORT'

# Octavia providers
AMPHORA_PROVIDER = 'amphora'
OVN_PROVIDER = 'ovn'

# Octavia services
WORKER_SERVICE = 'tripleo_octavia_worker.service'
HOUSEKEEPING_SERVICE = 'tripleo_octavia_housekeeping.service'
HM_SERVICE = 'tripleo_octavia_health_manager.service'
API_SERVICE = 'tripleo_octavia_api.service'

# Octavia containers
WORKER_CONTAINER = 'octavia_worker'
HOUSEKEEPING_CONTAINER = 'octavia_housekeeping'
HM_CONTAINER = 'octavia_health_manager'
API_CONTAINER = 'octavia_api'

# Octavia amphora provider resources names
LB_AMP_NAME = 'tobiko_octavia_amphora_lb'
LISTENER_AMP_NAME = 'tobiko_octavia_http_listener'
POOL_AMP_NAME = 'tobiko_octavia_http_pool'
MEMBER_AMP_NAME_PREFIX = 'tobiko_octavia_http_member'

# Octavia ovn provider resources names
LB_OVN_NAME = 'tobiko_octavia_ovn_lb'
LISTENER_OVN_NAME = 'tobiko_octavia_tcp_listener'
POOL_OVN_NAME = 'tobiko_octavia_tcp_pool'
MEMBER_OVN_NAME_PREFIX = 'tobiko_octavia_tcp_member'

# Providers/lb-names dictionary
OCTAVIA_PROVIDERS_NAMES = {
    'lb': {
        AMPHORA_PROVIDER: LB_AMP_NAME,
        OVN_PROVIDER: LB_OVN_NAME
    },
    'listener': {
        AMPHORA_PROVIDER: LISTENER_AMP_NAME,
        OVN_PROVIDER: LISTENER_OVN_NAME
    },
    'pool': {
        AMPHORA_PROVIDER: POOL_AMP_NAME,
        OVN_PROVIDER: POOL_OVN_NAME
    },
    'member': {
        AMPHORA_PROVIDER: MEMBER_AMP_NAME_PREFIX,
        OVN_PROVIDER: MEMBER_OVN_NAME_PREFIX
    }
}



# Copyright (c) 2023 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import openstack

import tobiko
from tobiko import config
from tobiko.openstack import keystone


CONF = config.CONF


class OpenstacksdkClientFixture(tobiko.SharedFixture):

    client = None

    def __init__(self, client=None):
        super(OpenstacksdkClientFixture, self).__init__()
        if client:
            self.client = client

    def setup_fixture(self):
        self.setup_client()

    def setup_client(self):
        client = self.client
        # create a new connection if it was not created before or if TLS-e is
        # enabled (otherwise, an SSLError exception is raised)
        if not client:
            credentials = keystone.keystone_credentials()
            tmp_auth = {
                'os-auth-url': credentials.auth_url,
                'os-password': credentials.password,
                'os-username': credentials.username,
                'os-cacert': credentials.cacert,
                'os-project-name': credentials.project_name,
                'os-user-domain-name': credentials.user_domain_name,
                'os-project-domain-name': credentials.project_domain_name,
                'os-project-domain-id': credentials.project_domain_id
            }
            if credentials.api_version == 3:
                tmp_auth['os-identity-api-version'] = credentials.api_version
            if 'https://' in credentials.auth_url and not credentials.cacert:
                tmp_auth['os-cacert'] = \
                    CONF.tobiko.tripleo.undercloud_cacert_file
            self.client = client = openstack.connect(**tmp_auth)
        return client


def openstacksdk_client():
    fixture = tobiko.setup_fixture(OpenstacksdkClientFixture())
    return fixture.client



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import itertools

from oslo_config import cfg

GROUP_NAME = 'neutron'
OPTIONS = [
    cfg.StrOpt('external_network',
               help="Network for creating ports on an external network",
               default="public"),
    cfg.StrOpt('ipv4_cidr',
               default='10.100.0.0/16',
               help="The CIDR block to allocate IPv4 subnets from"),
    cfg.IntOpt('ipv4_prefixlen',
               default=24,
               help="The mask bits for IPv4 subnets"),
    cfg.ListOpt('ipv4_dns_nameservers',
                # This value is a workaround to avoid UbuntuExternalPortTest
                # failures - with certain ubuntu images, routes to the
                # nameservers were added to a VLAN interface, breaking
                # connectivity from the VMs to the external network.
                default=["8.8.8.8"],
                help="List of nameservers IPv4 addresses"),
    cfg.StrOpt('ipv6_cidr',
               default='fc00::/48',
               help="The CIDR block to allocate IPv6 subnets from"),
    cfg.IntOpt('ipv6_prefixlen',
               default=64,
               help="The mask bits for IPv6 subnets"),
    cfg.ListOpt('ipv6_dns_nameservers',
                default=None,
                help="List of nameservers IPv6 addresses"),
    cfg.IntOpt('custom_mtu_size',
               default=1300,
               help=("Customized maximum transfer unit size\n"
                     "Notes:\n"
                     " - MTU values as small as 1000 has been seen "
                     "breaking networking binding due to an "
                     "unknown cause.\n"
                     " - Too big MTU values (like greater than 1400)"
                     " may be refused during network creation")),
    cfg.StrOpt('nameservers_host',
               default=None,
               help="Host where nameservers files are located"),
    cfg.ListOpt('nameservers_filenames',
                default=['/etc/resolv.conf'],
                help="File to parse for getting default nameservers list"),
    cfg.IntOpt('bwlimit_kbps',
               default=1000,
               help="The BW limit value configured for the QoS Policy Rule"),
    cfg.StrOpt('direction',
               default='egress',
               help="The direction for the QoS Policy Rule"),
    cfg.IntOpt('dscp_mark',
               default=40,
               help="The DSCP marking value for the QoS Policy Rule"),
    cfg.IntOpt('vlan_id',
               default=101,
               help="VLAN trunk subport segmentation ID"),
    cfg.ListOpt('ignore_extensions',
                default=[],
                help="Behave as these networking extensions are not available"
                ),
]


def register_tobiko_options(conf):
    conf.register_opts(group=cfg.OptGroup(GROUP_NAME), opts=OPTIONS)


def list_options():
    return [(GROUP_NAME, itertools.chain(OPTIONS))]



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import
from __future__ import division

import json
import typing

import netaddr
from oslo_log import log

import tobiko
from tobiko.shell.iperf3 import _interface
from tobiko.shell.iperf3 import _parameters
from tobiko.shell import sh
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


def get_bandwidth(address: typing.Union[str, netaddr.IPAddress],
                  bitrate: int = None,
                  download: bool = None,
                  port: int = None,
                  protocol: str = None,
                  ssh_client: ssh.SSHClientType = None,
                  timeout: tobiko.Seconds = None) -> float:
    iperf_measures = execute_iperf3_client(address=address,
                                           bitrate=bitrate,
                                           download=download,
                                           port=port,
                                           protocol=protocol,
                                           ssh_client=ssh_client,
                                           timeout=timeout)
    return calculate_bandwith(iperf_measures)


def calculate_bandwith(iperf_measures) -> float:
    # first interval is removed because BW measured during it is not
    # limited - it takes ~ 1 second to traffic shaping algorithm to apply
    # bw limit properly (buffer is empty when traffic starts being sent)
    intervals = iperf_measures['intervals'][1:]
    bits_received = sum([interval['sum']['bytes'] * 8
                         for interval in intervals])
    elapsed_time = sum([interval['sum']['seconds']
                        for interval in intervals])
    # bw in bits per second
    return bits_received / elapsed_time


def execute_iperf3_client(address: typing.Union[str, netaddr.IPAddress],
                          bitrate: int = None,
                          download: bool = None,
                          port: int = None,
                          protocol: str = None,
                          ssh_client: ssh.SSHClientType = None,
                          timeout: tobiko.Seconds = None) \
        -> typing.Dict:
    params_timeout: typing.Optional[int] = None
    if timeout is not None:
        params_timeout = int(timeout - 0.5)
    parameters = _parameters.iperf3_client_parameters(address=address,
                                                      bitrate=bitrate,
                                                      download=download,
                                                      port=port,
                                                      protocol=protocol,
                                                      timeout=params_timeout)
    command = _interface.get_iperf3_client_command(parameters)

    # output is a dictionary
    output = sh.execute(command,
                        ssh_client=ssh_client,
                        timeout=timeout).stdout
    return json.loads(output)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import inspect

import tobiko
from tobiko.openstack.keystone import _client


class ServiceListFixture(tobiko.SharedFixture):

    client = None
    services = None

    def setup_fixture(self):
        self.services = _client.list_services()

    def has_service(self, **attributes):
        services = self.services
        if services and attributes:
            services = services.with_attributes(**attributes)
        return bool(services)


def has_service(**attributes):
    fixture = tobiko.setup_fixture(ServiceListFixture)
    return fixture.has_service(**attributes)


def is_service_missing(**params):
    # return False if it is called from a unit test
    test_module_name = inspect.getmodule(tobiko.get_test_case()).__name__
    return (not test_module_name.startswith('tobiko.tests.unit.') and
            not has_service(**params))


def skip_if_missing_service(**params):
    return tobiko.skip_if('missing service: {!r}'.format(params),
                          is_service_missing, **params)



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko
from tobiko.shell import ssh
from tobiko.shell.sh import _execute


def make_temp_dir(ssh_client: ssh.SSHClientType = None,
                  sudo: bool = None) \
        -> str:
    test_case = tobiko.get_test_case()
    dir_name: str = _execute.execute('mktemp -d',
                                     ssh_client=ssh_client,
                                     sudo=sudo).stdout.strip()
    test_case.addCleanup(_execute.execute,
                         f'rm -fR "{dir_name}"',
                         ssh_client=ssh_client,
                         sudo=sudo)
    return dir_name



# Copyright (c) 2022 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import itertools
import os
import socket

import testtools

import tobiko
from tobiko import BackgroundProcessFixture


def get_sock_file() -> str:
    return tobiko.tobiko_config_path(f'~/.tobiko/tests/{__name__}/sock')


class MyBackgroundProcessFixture(BackgroundProcessFixture):

    def run(self, *args, **kwargs):
        self.serve_numbers()

    def serve_numbers(self):
        sock_file = get_sock_file()
        os.makedirs(os.path.dirname(sock_file), exist_ok=True)
        try:
            os.unlink(sock_file)
        except OSError:
            if os.path.exists(sock_file):
                raise
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            sock.bind(sock_file)
            sock.listen(1)
            for i in itertools.count():
                connection, _client_address = sock.accept()
                try:
                    message = bytes(f'{i}', encoding='utf-8')
                    connection.send(message)
                except Exception as ex:
                    print(ex)
                finally:
                    connection.close()
        finally:
            sock.close()

    def request_number(self, timeout=30.) -> int:
        connection = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        with connection:
            for attempt in tobiko.retry(timeout=timeout):
                try:
                    connection.connect(get_sock_file())
                    break
                except (ConnectionRefusedError, FileNotFoundError) as ex:
                    if attempt.is_last:
                        raise RuntimeError('Server not running') from ex
            message = connection.recv(4096)
            return int(message)


class BackgroundProcessTest(testtools.TestCase):

    process = tobiko.required_fixture(MyBackgroundProcessFixture,
                                      setup=False)

    def test_start(self):
        self.stop_process()
        number0 = self.start_process()
        self.assertLess(number0, self.start_process(),
                        "process has been restarted")

    def test_stop(self):
        number0 = self.start_process()
        self.stop_process()
        self.assertGreaterEqual(number0, self.start_process(),
                                "process not stopped")

    def start_process(self) -> int:
        # pylint: disable=protected-access
        self.process.start()
        self.assertTrue(self.process.is_alive)
        self.assertTrue(os.path.isfile(self.process._pid_file))
        return self.process.request_number()

    def stop_process(self):
        # pylint: disable=protected-access
        self.process.stop()
        self.assertFalse(self.process.is_alive)
        self.assertFalse(os.path.isfile(self.process._pid_file))
        self.assertRaises(RuntimeError, self.process.request_number,
                          timeout=5.)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import typing

import testtools

from tobiko import run


class FindTestFilesTest(testtools.TestCase):

    def test_find_test_files(self,
                             test_path: typing.List[str] = None):
        if test_path is None:
            test_path = [__file__]
        test_files = run.find_test_files(test_path=test_path)
        self.assertIn(__file__, test_files)
        test_path = [os.path.realpath(path)
                     for path in test_path]
        for test_file in test_files:
            for path in test_path:
                if test_file.startswith(path):
                    break
            else:
                self.fail(f"File '{test_file}' not in any path ({test_path})")

    def test_find_test_files_with_test_dir(self):
        return self.test_find_test_files(
            test_path=[os.path.dirname(__file__)])
